Package: adverSCarial
Title: adverSCarial, generate and analyze the vulnerability of
        scRNA-seq classifier to adversarial attacks
Version: 1.11.0
Authors@R: c(
    person("Ghislain", "FIEVET", , "ghislain.fievet@gmail.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-0337-7327")),
    person("Sébastien", "HERGALANT", role = "aut",
    email="sebastien.hergalant@inserm.fr", comment=c(ORCID = "0000-0001-8456-7992")))
Description: 
    adverSCarial is an R Package designed for generating and analyzing the vulnerability of scRNA-seq
    classifiers to adversarial attacks. The package is versatile and provides a format for integrating
    any type of classifier. It offers functions for studying and generating two types of attacks,
    single gene attack and max change attack. The single-gene attack involves making a small modification
    to the input to alter the classification. The max-change attack involves making a large modification
    to the input without changing its classification. The CGD attack is based on an estimated gradient descent.
    against adversarial attacks.
    The package provides a comprehensive solution for evaluating the robustness of scRNA-seq classifiers
    against adversarial attacks.
License: MIT + file LICENSE
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.2.3
biocViews: Software, SingleCell, Transcriptomics, Classification
Suggests: knitr, RUnit, BiocGenerics, TENxPBMCData, CHETAH, stringr,
        LoomExperiment
Imports: gtools, S4Vectors, methods, DelayedArray
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/adverSCarial
git_branch: devel
git_last_commit: 0818fc6
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 21:10:19 UTC; biocbuild
Author: Ghislain FIEVET [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-0337-7327>),
  Sébastien HERGALANT [aut] (ORCID:
    <https://orcid.org/0000-0001-8456-7992>)
Maintainer: Ghislain FIEVET <ghislain.fievet@gmail.com>
Depends: R (>= 3.5.0)

# Generated by roxygen2: do not edit by hand

export(MClassifier)
export(advCGD)
export(advChar)
export(advGridMinChange)
export(advList)
export(advMaxChange)
export(advModifications)
export(advRandWalkMinChange)
export(advSingleGene)
export(advTraining)
export(getSignGenes)
export(matrixFromSCE)
export(maxChangeOverview)
export(predictWithNewValue)
export(sceConvertToHGNC)
export(singleGeneOverview)
exportClasses(advChar)
exportClasses(advList)
importFrom(methods,setMethod)
importFrom(methods,setClass)

#' Returns a modified RNA expression DelayedMatrix, or a modified SingleCellExperiment,
#' for a given cluster, for a given modification.
#'
#' @details The motivation for this function is to standardize the modifications
#' we want to study in the attacks. We give as argument a DelayedMatrix of the RNA
#' expression, the gene and the target cells we want to modify. Then we have
#' three arguments allowing to specify what modification we want to apply on
#' these cells. The advMethod contains, a specific prerecorded modification
#' or an indication on how to use the other two arguments.
#' The prerecorded modifications available for the advMethod argument are:
#' - 'perc1', replace the value by the whole matrix 1 percentile value of
#'  the gene. It is as if we biologically switched off the gene.
#' - 'perc99', replace the value by the whole matrix 99 percentile value of
#'  the gene. It is as if we biologically switched on the gene to the maximum.
#' - 'random', replace the value by from a uniform distribution between min 
#' and max of the gene on the dataset
#' - 'positive_aberrant' replace value by 10,000 times the max value of the
#' gene on the dataset
#' - 'negative_aberrant' replace value by -10,000 times the max value of the
#' gene on the dataset
#' - 'decile+X', shifts the gene value by + X deciles.
#' - 'decile-X', shifts the gene value by - X deciles.
#' The value of the advMethod argument can also be 'fixed', in this case the
#' modification would be to replace the value of the gene of the wanted cells
#' by the value of the argument 'advFixedValue'. This can be useful to test
#' aberrant values like negative integer, absurdly high values of character
#' values.
#' The value of the advMethod argument can also be 'full_row_fct',
#' 'target_row_fct', 'target_matrix_fct' or 'full_matrix_fct'. They are
#' used when we want to use a custom modification function, with the 'advFct'
#' argument:
#' - 'full_row_fct' indicate that the 'advFct' function takes the whole gene
#'  values as input.
#' - 'target_row_fct' indicate that the 'advFct' function takes target cells
#'  gene values as input.
#' - 'full_matrix_fct' indicate that the 'advFct' function takes the whole gene
#'  expression values as input.
#' - 'target_matrix_fct' indicate that the 'advFct' function takes target cells
#'  all genes values as input.
#' 
#' @param exprs DelayedMatrix of numeric RNA expression, cells are rows and genes
#' are columns - or a SingleCellExperiment object, a matrix or a data.frame. By default,
#' these are converted to a data.frame to increase speed performance during modifications.
#' However, this conversion can consume a significant amount of memory, see 'argForModif'
#' argument for options.
#' @param genes the character vector of genes to modify
#' @param clusters a character vector of the clusters to which the cells belong
#' @param target the name of the cluster to modify
#' @param advMethod the name of the method to use
#' @param advFixedValue the numeric value to use in case of
#' advMethod=`fixed`
#' @param advFct the function to use in case advMethod
#' belongs to the following list: `full_row_fct`, `target_row_fct`,
#' `target_matrix_fct`, `full_matrix_fct`
#' @param argForClassif the type of the first argument to feed to the
#' classifier function. 'DelayedMatrix' by default, can be 'SingleCellExperiment'
#' or 'data.frame'.
#' @param argForModif type of matrix during for the modification, 'DelayedMatrix'
#' by default. Can be 'data.frame', which is faster, but need more memory.
#' @param verbose logical, set to TRUE to activate verbose mode
#' @return the matrix or a data.frame exprs modified on asked genes
#' with the specified modification
#' @examples
#' library(DelayedArray)
#' 
#' rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
#'      CD8B=c(2,2,3,3)))
#' genes <- c("CD4", "CD8A")
#' clusters_id <- c("B cell","B cell","T cell","T cell")
#'
#' advModifications(rna_expression, genes, clusters_id,
#' "T cell", advMethod="perc99")
#' @export
advModifications <- function(exprs, genes, clusters,
                            target, advMethod = "perc99",
                            advFixedValue = 3, advFct = NULL,
                            argForClassif = 'DelayedMatrix',
                            argForModif = 'data.frame',
                            slot=NULL,
                            verbose = FALSE) {
    if (!is(exprs, 'matrix') && !is(exprs,'data.frame') &&
        !is(exprs,'SingleCellExperiment') && !is(exprs,'DelayedMatrix') &&
        !is(exprs,'Seurat')){
        stop("The argument exprs must be a DelayedMatrix, a SingleCellExperiment, a matrix or a data.frame")
    }
    if (!is.character(genes)) {
        stop("The argument genes must be character or vector of character.")
    }
    if (!is.character(clusters)) {
        stop("The argument clusters must be a vector of character.")
    }
    if (!is.character(target)) {
        stop("The argument target must be character.")
    }
    if (!is.character(advMethod)) {
        stop("The argument advMethod must be character.")
    }
    if (!is.character(argForClassif)) {
        stop("The argument argForClassif must be character: 'data.frame' or 'SingleCellExperiment'.")
    }
    if (!is.logical(verbose)){
        stop("The argument verbose must be logical.")
    }
    if (verbose) {
        message("Modify data for ", length(genes), " genes for cluster ", target)
    }
    # if (is(exprs,'Seurat') && tolower(argForModif)=="seurat"){
    if (is(exprs,'Seurat')){
        so_exprs <- exprs
        exprs <- as.data.frame(so_exprs@assays$RNA@layers[[slot]])
        colnames(exprs) <- rownames(so_exprs@assays$RNA@cells)
        rownames(exprs) <- rownames(so_exprs@assays$RNA@features)
        exprs <- as.data.frame(t(exprs))
    }
    if (is(exprs,'SingleCellExperiment') ){
        exprs <- t(as.matrix(counts(exprs)))
    }
    if (!is(exprs,'DelayedMatrix') && argForModif=="DelayedMatrix"){
        if (verbose){
            message("Converting exprs object to a DelayedArray object")
        }
        exprs <- DelayedArray::DelayedArray(exprs)
    }


    if (!is(exprs,'data.frame') && argForModif=="data.frame"){
        if (verbose){
            message("Converting exprs object to a data.frame object")
        }
        exprs <- as.data.frame(exprs)
    }

    
    if (advMethod != "none"){
        if (advMethod %in% c("random", "positive_aberrant", "negative_aberrant")){
            if (advMethod == "random"){
                exprs <- .advModificationsFunction(exprs, genes, clusters,
                                    target, advMethod = "full_matrix_fct",
                                    advFct = .fctRand)
            }
            if (advMethod == "positive_aberrant"){
                exprs <- .advModificationsFunction(exprs, genes, clusters,
                                    target, advMethod = "full_row_fct",
                                    advFct = .fctAbbPos)
            }
            if (advMethod == "negative_aberrant"){
                exprs <- .advModificationsFunction(exprs, genes, clusters,
                                    target, advMethod = "full_row_fct",
                                    advFct = .fctAbbNeg)
            }
        } else {
            if (startsWith(advMethod, "decile")){
                if ( unlist(strsplit(advMethod, "\\+"))[1] == "decile" ){
                    decilNumber <- as.numeric(unlist(strsplit(advMethod, "\\+")))[2]
                    exprs <- .advModificationsFunction(exprs, genes, clusters,
                                target, advMethod = "full_row_fct",
                                advFct = function(x,y){.advDecil(x,y,decilNumber=decilNumber)})
                }
                if ( unlist(strsplit(advMethod, "\\-"))[1] == "decile" ){
                    decilNumber <- - as.numeric(unlist(strsplit(advMethod, "\\-")))[2]
                    exprs <- .advModificationsFunction(exprs, genes, clusters,
                                target, advMethod = "full_row_fct",
                                advFct = function(x,y){.advDecil(x,y,decilNumber=decilNumber)})
                }
            } else {
                if (!is.function(advFct)) {
                    exprs <- .advModificationsNotFunction(exprs,
                                                genes, clusters,
                                                target, advMethod = advMethod,
                                                advFixedValue = advFixedValue)
                } else {
                    exprs <- .advModificationsFunction(exprs, genes, clusters,
                                        target, advMethod = advMethod,
                                        advFct = advFct)
                }
            }
        }


        if(tolower(argForClassif)=="seurat"){
            so_exprs@assays$RNA@layers[[slot]] <- t(exprs)
            exprs <- so_exprs
        }
    } else {
        if(tolower(argForClassif)=="seurat"){
            exprs <- so_exprs
        }
    }

    if ( argForClassif == 'SingleCellExperiment'){
        exprs <- SingleCellExperiment(assays = list(counts = t(exprs)))
    }
    if (!is(exprs, 'data.frame') && argForClassif=='data.frame'){
        exprs <- as.data.frame(exprs)
    }
    if (!is(exprs,'DelayedMatrix') && argForClassif=="DelayedMatrix"){
        exprs <- DelayedArray::DelayedArray(exprs)
    }

    exprs
}

# Function for the default "random" modification
.fctRand <- function(x, cMask){
    sample(min(x):max(x), sum(cMask), replace=T)
}

# Function for the default "positive_aberrant" modification
.fctAbbPos <- function(x, y){10000*max(x)}

# Function for the default "negative_aberrant" modification
.fctAbbNeg <- function(x, y){-10000*max(x)}

.advModificationsNotFunction <- function(exprs,
                            genes, clusters,
                            target, advMethod = "perc99",
                            advFixedValue = 3) {
    if (is(exprs,'SingleCellExperiment')){
        exprs <- t(as.matrix(counts(exprs)))
    }
    cellMask <- clusters == target
    if (advMethod == "fixed") {
        numGenes <- unlist(lapply(unique(genes), function(myGene){
            if (is(exprs[, myGene], "numeric")) {
                return(myGene)
            }
        }))
        exprs[cellMask, numGenes] <- as.numeric(advFixedValue)
    }
    if (advMethod == "perc99") {
        myGenes <- unlist(lapply(unique(genes), function(myGene){
            if(is(exprs[, myGene], "numeric")){
                return(myGene)
            } else {
                return(NULL)
            }
        }))
        myQuantiles <- unlist(lapply(myGenes, function(myGene){
            stats::quantile(exprs[, myGene], 0.99)
        }))
        exprs[cellMask, myGenes] <- rep(myQuantiles, each=sum(cellMask))
    }
    if (advMethod == "perc1") {
        myGenes <- unlist(lapply(unique(genes), function(myGene){
            if(is(exprs[, myGene], "numeric")){
                return(myGene)
            } else {
                return(NULL)
            }
        }))
        myQuantiles <- unlist(lapply(myGenes, function(myGene){
            stats::quantile(exprs[, myGene], 0.01)
        }))
        exprs[cellMask, myGenes] <- rep(myQuantiles, each=sum(cellMask))
    }
    exprs
}

.advModificationsFunction <- function(exprs, genes, clusters,
                            target, advMethod = "perc99",
                            advFct = NULL) {
    cellMask <- clusters == target
    if (advMethod == "full_row_fct" || advMethod == "fixed") {
        for (myGene in unique(genes)) {
            if (is(exprs[, myGene], "numeric")) {
                exprs[cellMask, myGene] <- advFct(exprs[, myGene], cellMask)
            }
        }
    }
    if (advMethod == "target_row_fct" || advMethod == "fixed") {
        for (myGene in unique(genes)) {
            if (is(exprs[, myGene], "numeric")) {
                exprs[cellMask, myGene] <- advFct(exprs[
                    cellMask,
                    myGene
                ])
            }
        }
    }
    if (advMethod == "target_matrix_fct") {
        numGenes <- unlist(lapply(unique(genes),function(myGene){
            if (is(exprs[, myGene], "numeric")) {
                return(myGene)
            } else {
                return(NULL)
            }
        }))
        exprs[cellMask, numGenes] <- advFct(exprs[cellMask, numGenes])
    }
    if (advMethod == "full_matrix_fct") {
        numGenes <- unlist(lapply(unique(genes),function(myGene){
            if (is(exprs[, myGene], "numeric")) {
                return(myGene)
            } else {
                return(NULL)
            }
        }))
        exprs[cellMask, numGenes] <- advFct(exprs[, numGenes], cellMask)
    }
    exprs
}

.advDecil <- function(x, y, decilNumber=1){
    orderedX <- x[order(x)]
    modValues <- unlist(lapply(x[y], function(val){
        xRank <- sum(orderedX < val)
        if ( decilNumber > 0){
            if ( xRank + round(decilNumber*length(x) / 10) > length(x) ){
                return(max(x))
            } else {
                return(orderedX[xRank + round(decilNumber*length(x) / 10)])
            }
        } else {
            if ( xRank + round(decilNumber*length(x) / 10) < 1 ){
                return(min(x))
            } else {
                return(orderedX[xRank + round(decilNumber*length(x) / 10)])
            }
        }
    }))
    modValues
}

library(RUnit)
library(adverSCarial)

runTests <- function() {
    test_advModifications()
    test_predictWithNewValue()
    test_advMaxChange()
    test_advSingleGene()
    test_maxChangeOverview()
    test_singleGeneOverview()
    test_advGridMinChange()
    test_advRandWalkMinChange()
}

test_advModifications <- function () {
    exprs <- data.frame(a=c(0,0,0,0), b=c(1,1,1,1), c=c(2,2,3,3))
    genes <- c("a")
    clusters <- c("b","b","t","t")
    exprs_res1 <- data.frame(a=c(10,10,0,0), b=c(1,1,1,1), c=c(2,2,3,3))
    checkTrue( all(advModifications(exprs, genes, clusters, "b",
        advMethod = "fixed",
        advFixedValue = 10, argForClassif='data.frame') == exprs_res1))
    checkTrue(all(advModifications(
        data.frame(a=c(1)), "a", "a", "a") == 1))
}



MyClassifier <- function(expr, clusters, target) {
    c("b", 0.5)
}

test_predictWithNewValue <- function(){
    exprs <- data.frame(a=c(0,0,0,0), b=c(1,1,1,1), c=c(2,2,3,3))
    genes <- c("a")
    clusters <- c("b","b","t","t")
    checkTrue(all(predictWithNewValue(exprs, genes, clusters, "b",
                                MyClassifier) == c("b", 0.5)))
}

test_advMaxChange <- function(){
    exprs <- data.frame(a=c(0,0,0,0), b=c(1,1,1,1), c=c(2,2,3,3))
    clusters <- c("b","b","t","t")
    checkTrue(length(advMaxChange(exprs, clusters, "b", MyClassifier)@values) == 3)
}

test_advSingleGene <- function(){
    exprs <- data.frame(a=c(0,0,0,0), b=c(1,1,1,1), c=c(2,2,3,3))
    clusters <- c("b","b","t","t")
    checkTrue(length(advSingleGene(exprs, clusters, "b", MyClassifier)@values) == 0)
}

test_maxChangeOverview <- function(){
    exprs <- data.frame(a=c(0,0,0,0), b=c(1,1,1,1), c=c(2,2,3,3))
    clusters <- c("b","b","t","t")
    checkTrue(all(dim(maxChangeOverview(exprs, clusters, MyClassifier))==c(2,2)))
}

test_singleGeneOverview <- function(){
    exprs <- data.frame(a=c(0,0,0,0), b=c(1,1,1,1), c=c(2,2,3,3))
    clusters <- c("b","b","t","t")
    checkTrue(all(dim(singleGeneOverview(exprs, clusters, MyClassifier))==c(2,2)))
}

test_advGridMinChange <- function(){
    exprs <- data.frame(a=c(0,0,0,0), b=c(1,1,1,1), c=c(2,2,3,3))
    genes <- c("a")
    clusters <- c("b","b","t","t")

    checkTrue(all(dim(advGridMinChange(exprs, clusters, "b", MyClassifier,
                genes)) == c(3,5)))
}

test_advRandWalkMinChange <- function(){
    exprs <- data.frame(a=c(0,0,0,0), b=c(1,1,1,1), c=c(2,2,3,3))
    genes <- c("a")
    clusters <- c("b","b","t","t")

    checkTrue(all(dim(advRandWalkMinChange(exprs,
        clusters, "b", MyClassifier, genes)) == c(3,6)))
}

#' Returns a SingleCellExperiment object keeping unique HGNC gene
#'
#' @details Sometimes classifiers need HGNC instead of ensemble genes
#' to run. This function allows to make the conversion.
#' @param sce SingleCellExperiment object to convert
#' @return the SingleCellExperiment object keeping unique HGNC gene
#' @examples
#' library(TENxPBMCData)
#'
#' pbmc <- TENxPBMCData(dataset = "pbmc3k")
#' hgnc_pbmc <- sceConvertToHGNC(pbmc)
#'
#' @export
sceConvertToHGNC <- function(sce){
    if ( !is(sce,'SingleCellExperiment')){
        stop("The argument sce must be a SingleCellExperiment.")
    }
    matSce <- matrixFromSCE(sce)
    SingleCellExperiment(assays = list(counts = t(matSce)), colData = colData(sce))
}

#' @export
advTraining <- function(expr, classifier, genes=NULL,
                    coef=0.1, add=0, slot=NULL, geneNumber=100,
                    modGeneNumber=10000000, verbose=FALSE){
    if (class(expr) == "Seurat"){
        cell_number <- nrow(expr@assays$RNA@cells)
    } else {
        cell_number <- nrow(expr)
    }
    predictions <- classifier(expr, rep("all", cell_number), "all")
    cellPredictions <- predictions$cellTypes

    if (is.null(genes) && class(expr) != "Seurat"){
        genes <- colnames(expr)[sample(1:ncol(expr), ncol(expr))]
    }
    if ( is.null(genes) && class(expr) == "Seurat"){
        genes <- expr@assays$RNA@features[sample(1:length(expr@assays$RNA@features), length(expr@assays$RNA@features))]
   }

    modExpr <- expr
    tempExpr <- expr
    geneCount <- 1
    modGeneCount <- 0
    geneInDS <- TRUE
    modifiedGenes <- c()
    while(geneCount <= geneNumber && modGeneCount <= modGeneNumber){
        
        myGene <- genes[geneCount]
        if(verbose){
            message(myGene, " ", geneCount)
        }
    
        if (class(expr) == "Seurat"){
            dfScaled <- as.data.frame(modExpr@assays$RNA@layers[[slot]])
            colnames(dfScaled) <- rownames(modExpr@assays$RNA@cells)
            rownames(dfScaled) <- rownames(modExpr@assays$RNA@features)
            dfScaled <- as.data.frame(t(dfScaled))
            # modify only cells out of target cells
            if ( !myGene %in% colnames(dfScaled)){
                myGene <- sub( "\\.","-",myGene)
                if (!myGene %in% colnames(dfScaled)){
                    geneInDS <- FALSE
                }
            }
            if (geneInDS){
                dfScaled[, myGene][dfScaled[, myGene]>0] <- dfScaled[, myGene][dfScaled[, myGene]>0]*(1+coef) + add
                dfScaled[, myGene][dfScaled[, myGene]<= 0] <- dfScaled[, myGene][dfScaled[, myGene] <= 0]*(1-coef) + add
                modExpr@assays$RNA@layers[[slot]] <- t(dfScaled)
            }
        } else {
            # modExpr[, myGene] <- modExpr[, myGene]*(1+coef) + add    
            modExpr[, myGene][modExpr[, myGene]>0] <- modExpr[, myGene][modExpr[, myGene]>0]*(1+coef) + add
            modExpr[, myGene][modExpr[, myGene]<= 0] <- modExpr[, myGene][modExpr[, myGene] <= 0]*(1-coef) + add
        }
        
        newPredictions <- classifier(modExpr, rep("all", cell_number), "all")
        newPredictions$cellTypes
        newPredictions$typePredictions

        dfnp <- t(newPredictions$typePredictions[,colnames(predictions$typePredictions)])
        dfp <- t(predictions$typePredictions)
        allCols <- unique(c(colnames(dfnp), colnames(dfp)))
        dfnp[,setdiff(allCols, colnames(dfnp))] <- 0
        dfp[,setdiff(allCols, colnames(dfp))] <- 0
        dfnp <- dfnp[, order(colnames(dfnp))]
        dfp <- dfp[,colnames(dfnp)]
        
        # Make sure colnames are the same, same length
        colnames(dfnp) <- paste0("new_",colnames(dfnp))
        colnames(dfp) <-paste0("or_",colnames(dfp))
        df <- cbind(dfnp, dfp)
        scoreDelts <- t(apply(df, 1, function(x){
            newVals <- x[grep("^new_", names(x), value=TRUE)]
            orVals <- x[grep("^or_", names(x), value=TRUE)]
            newVals - orVals
        }))
        
        # Which cells benefit from modif
        scoreDelts <- as.data.frame(scoreDelts)
        scoreDelts$cell_id <- rownames(scoreDelts)
        scoreDelts$prevCellPredictions <- cellPredictions

        # Which cells benefit from modif
        cbfm <- unlist(apply(scoreDelts, 1, function(x){
            y <- x[order(x, decreasing = FALSE)]
            cellOrTarget <- names(y)[1]
            cellNewTarget <- names(y)[2]
            if (as.numeric(unname(x[cellNewTarget])) > as.numeric(unname(x[cellOrTarget]))) {
                return(x["cell_id"])
            }
        }))
        # Which cells benefit from opposit modif
        cbfom <- unlist(apply(scoreDelts, 1, function(x){
            y <- x[order(x, decreasing = FALSE)]
            cellOrTarget <- names(y)[1]
            cellNewTarget <- names(y)[2]
            if (as.numeric(unname(x[cellNewTarget])) < as.numeric(unname(x[cellOrTarget]))) {
                return(x["cell_id"])
            }
        }))

        modExpr <- tempExpr

        if ((length(cbfm) + length(cbfom)) > 0){
            modifiedGenes <- c(modifiedGenes, myGene)
            if (class(expr) == "Seurat"){
                dfScaled <- as.data.frame(modExpr@assays$RNA@layers[[slot]])
                colnames(dfScaled) <- rownames(modExpr@assays$RNA@cells)
                rownames(dfScaled) <- rownames(modExpr@assays$RNA@features)
                dfScaled <- as.data.frame(t(dfScaled))
                dfScaled[cbfm, myGene][dfScaled[cbfm, myGene]>0] <- dfScaled[cbfm, myGene][dfScaled[cbfm, myGene]>0]*(1+coef) + add
                dfScaled[cbfm, myGene][dfScaled[cbfm, myGene]<= 0] <- dfScaled[cbfm, myGene][dfScaled[cbfm, myGene] <= 0]*(1-coef) + add
#                 dfScaled[cbfm, myGene] <- dfScaled[cbfm, myGene]*(1+coef) + add
                dfScaled[cbfom, myGene][dfScaled[cbfom, myGene]>0] <- dfScaled[cbfom, myGene][dfScaled[cbfom, myGene]>0]*(1-coef) - add
                dfScaled[cbfom, myGene][dfScaled[cbfom, myGene]<= 0] <- dfScaled[cbfom, myGene][dfScaled[cbfom, myGene] <= 0]*(1+coef) - add
#                 dfScaled[cbfom, myGene] <- dfScaled[cbfom, myGene]*(1-coef) - add
                modExpr@assays$RNA@layers[[slot]] <- t(dfScaled)
                modGeneCount <- modGeneCount + 1
            } else {
                modExpr[cbfm, myGene] <- modExpr[cbfm, myGene]*(1+coef) + add
                modExpr[cbfom, myGene] <- modExpr[cbfom, myGene]*(1-coef) - add
            }
            tempExpr <- modExpr
        }
        geneCount <- geneCount + 1
    }
    return(modExpr)
}

#' Random walk search of min change adversarial attack.
#' 
#' @details The parameter search by grid can take a long time,
#' this function aims to make a parameter search faster.
#' We have a function, advSingleGene, looking for one gene attacks.
#' The advRandWalkMinChange function aims to find a multiple genes
#' attack, with the fewer genes possible.
#' At first the user have to provide a list of genes to test, for
#' example by running differential statistics between two cell clusters.
#' The user should also provide a list of modifications to test, to
#' define as so - each item of the list should be 1 or 2 length size.
#' The 1 length vector must contain the prerecorded modifications, 'perc1'
#' or 'perc99'.
#' The 2 length vector must have as first item:
#'  - 'fixed', in this case the second item should be the value to be
#'  replaced by.
#'  - 'full_row_fct', 'target_row_fct', 'target_matrix_fct' or
#'  'full_matrix_fct'. In this case the second item should be a function.
#' Let's say we want to analysis the susceptibility to min change attack
#' for 3 modifications: "perc1", the modification of each value of the
#' cluster by 1000, and a custom modification stored inside a function myFct.
#' Then the 'modification' parameter should be:
#' my_modifications = list(c("perc1"),
#'                         c("fixed", 1000),
#'                         c("full_matrix_fct", myFct))
#' 
#' Then the function will try to find the best combination of these genes
#' and modifications to make the min change attack.
#' Step 1 is to find a seed by trying random combinations of
#' genes and modifications on a cluster until the classification is altered.
#' Step 2 is to perform a random walk search to reduce the number of genes
#' needed to change the classification.
#' The 
#' @param exprs DelayedMatrix of numeric RNA expression, cells are rows and genes
#' are columns - or a SingleCellExperiment object, a matrix or a data.frame. By
#' default matrix and data.frame are converted to DelayedMatrix for memory
#' performance, see 'argForModif' argument for options.
#' @param clusters a character vector of the clusters to which the cells belong
#' @param target the name of the cluster to modify
#' @param classifier a classifier in the suitable format.
#' A classifier function should be formated as follow:
#' classifier = function(expr, clusters, target){
#'      # Making the classification
#'      c("cell type", score)
#' }
#' `score` should be numeric between 0 and 1, 1 being the highest confidance
#' into the cell type classification.
#' The matrix `expr` contains RNA expression values, the vector `clusters`
#' consists of the cluster IDs for each cell in `expr`, and `target` is the
#' ID of the cluster for which we want to have a classification.
#' The function returns a vector with the classification result, and a score.
#' @param genes the character vector of genes to study
#' @param modifications the list of the modifications to study
#' @param firstBatch the maximum number of try in step 1
#' @param walkLength the maximum number of try in step 2
#' @param stepChangeRatio ratio of parameters change in new walk step
#' @param whileMaxCount the maximum number of try when looking
#' for new combination of parameters
#' @param changeType `any` consider each misclassification,
#'  `not_na` consider each misclassification but NA.
#' @param argForClassif the type of the first argument to feed to the
#' classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
#' or 'DelayedMatrix'.
#' @param argForModif type of matrix during for the modification, 'data.frame'
#' by default. Can be 'DelayedMatrix', which needs less memory but is slower.
#' @param verbose logical, set to TRUE to activate verbose mode
#' @return DataFrame results of the classification of all the grid combinations
#' @examples
#' library(DelayedArray)
#' 
#' MyClassifier <- function(expr, clusters, target) {
#'    c("T cell", 0.9)
#' }
#' rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
#'      CD8B=c(2,2,3,3)))
#' genes <- c("CD4", "CD8A")
#' clusters_id <- c("B cell","B cell","T cell","T cell")
#'
#' advRandWalkMinChange(rna_expression, clusters_id, "T cell",
#' MyClassifier, genes=genes,
#' modifications = list(c("perc1"), c("perc99")))
#'
#' # Stop at first attack discovery, whitout going into the walk
#' # parameter search.
#' advRandWalkMinChange(rna_expression, clusters_id, "T cell",
#'  MyClassifier, genes=genes,
#'  modifications = list(c("perc1"), c("perc99")), walkLength=0)
#' 
#' myModif = function(x, y){
#'    return(sample(1:10,1))
#' }
#' 
#' my_modifications = list(c("perc1"),
#'                         c("fixed", 1000),
#'                         c("full_matrix_fct", myModif))
#' advRandWalkMinChange(rna_expression, clusters_id, "T cell",
#'  MyClassifier, genes=genes, modifications = my_modifications)
#' @export
advRandWalkMinChange <- function(exprs, clusters, target, classifier, genes,
        modifications = list(c("perc1"), c("perc99")), firstBatch = 100,
        walkLength = 100, stepChangeRatio = 0.2, whileMaxCount = 10000,
        changeType = "any", argForClassif = 'data.frame',
        argForModif = 'data.frame',
        verbose = FALSE) {
    if (!is(exprs, 'matrix') && !is(exprs,'data.frame') &&
        !is(exprs,'SingleCellExperiment') && !is(exprs,'DelayedMatrix')){
        stop("The argument exprs must be a DelayedMatrix, a SingleCellExperiment, a matrix or a data.frame")
    }
    if (!is.character(clusters)) {
        stop("The argument clusters must be a vector of character.")
    }
    if (!is.character(target)) {
        stop("The argument target must be character.")
    }
    if (!is.function(classifier)){
        stop("The argument classifier must be a function.")
    }
    if (!is.character(genes)) {
        stop("The argument genes must be character or vector of character.")
    }
    if (!is.list(modifications)) {
        stop("The argument modifications must be a list.")
    }
    if (!is.numeric(firstBatch)) {
        stop("The argument firstBatch must be numeric.")
    }
    if (!is.numeric(walkLength)) {
        stop("The argument walkLength must be numeric.")
    }
    if (!is.numeric(stepChangeRatio)) {
        stop("The argument stepChangeRatio must be numeric.")
    }
    if (!is.numeric(whileMaxCount)) {
        stop("The argument whileMaxCount must be numeric.")
    }
    if (!is.character(changeType)) {
        stop("The argument changeType must be character.")
    }
    if (!is.character(argForClassif)) {
        stop("The argument argForClassif must be character: 'data.frame' or 'SingleCellExperiment'.")
    }
    if (!is.logical(verbose)){
        stop("The argument verbose must be logical.")
    }
    if (is(exprs,'SingleCellExperiment') ){
        exprs <- t(as.matrix(counts(exprs)))
    }
    if (!is(exprs,'DelayedMatrix') && argForModif=="DelayedMatrix"){
        message("Converting exprs object to a DelayedArray object")
        exprs <- DelayedArray::DelayedArray(exprs)
    }
    if (!is(exprs,'data.frame') && argForModif=="data.frame"){
        message("Converting exprs object to a data.frame object")
        exprs <- as.data.frame(exprs)
    }

    beforeWalk <- .randWalkBeforeWalk(exprs, genes, modifications,
        clusters, target, classifier, firstBatch, argForClassif, argForModif, verbose)
    previousStrComb <- beforeWalk[[1]]
    functionResults <- beforeWalk[[2]]
    bestAttackParams <- beforeWalk[[3]]
    if (functionResults[1, "typeModified"] == FALSE) {
        message("No modified type, try with a higher firstBatch argument")
        return(functionResults)
    }
    for (i in seq_len(walkLength)) {
        if (verbose || log2(i) == round(log2(i)) || i == walkLength){
            message("Walk step ", i, " on ", walkLength)
            message("Current best attack length: ",
                length(bestAttackParams), " genes")
        }
        # Create a new vector of parameters to test
        rwNewVector <- .randWalkNewVector(previousStrComb, modifications,
            bestAttackParams, stepChangeRatio, whileMaxCount,
            functionResults)
        # If infinite loop kill function and return result
        if ( !rwNewVector[[1]] ) {return(rwNewVector[[2]])}
        newWalkParams <- rwNewVector[[2]]
        newWalkStep <- rwNewVector[[3]]

        previousStrComb <- c(previousStrComb, newWalkStep)
        lrwtnv <- .randWalkTryNewVector(exprs, genes,
            newWalkParams, modifications, clusters, target,
            classifier, bestAttackParams, i, functionResults,
            changeType, argForClassif, argForModif, verbose)
        functionResults <- lrwtnv[['functionResults']]
        bestAttackParams <- lrwtnv[['bestAttackParams']]
    }
    functionResults <- functionResults[order(as.numeric(functionResults$genesModified)), ]
    functionResults <- functionResults[order(functionResults$typeModified, decreasing = TRUE), ]
    S4Vectors::DataFrame(functionResults)
}

.randWalkTryNewVector <- function(exprs, genes, newWalkParams, modifications,
        clusters, target, classifier, bestAttackParams, i,
        functionResults, changeType, argForClassif, argForModif, verbose){
    exprsTemp <- exprs
    rowResults <- c()
    rowResultsInt <- c()
    # Try the new params
    for (geneInd in seq_along(genes)) {
        modifInd <- as.numeric(newWalkParams[geneInd])
        if (modifInd != length(modifications) + 1) {
            mod1 <- modifications[[modifInd]][[1]]
            if (length(modifications[[modifInd]]) == 1) {
                exprsTemp <- advModifications(exprsTemp,
                    genes[geneInd], clusters, target,
                    advMethod = mod1, argForModif=argForModif, verbose = verbose)
            } else {
                mod2 <- modifications[[modifInd]][[2]]
                exprsTemp <- advModifications(exprsTemp, genes[geneInd],
                    clusters, target, advMethod = mod1,
                    advFixedValue = mod2, advFct = mod2, argForModif=argForModif,
                    verbose = verbose)
            }
            rowResults <- c(rowResults, paste(modifications[[modifInd]],
                        collapse = " "))
            rowResultsInt <- c(rowResultsInt, modifInd)
        } else {
            rowResults <- c(rowResults, "NA")
            rowResultsInt <- c(rowResultsInt, modifInd)
        }
    }
    genesModified <- length(genes) - sum(rowResults == "NA")
    previousGenesModified <- length(genes) - sum(bestAttackParams == (length(modifications) + 1))
    if ( argForClassif == 'SingleCellExperiment'){
        exprsTemp <- SingleCellExperiment(assays = list(counts = t(exprsTemp)))
    }
    if (!is(exprsTemp, 'data.frame') && argForClassif=='data.frame'){
        exprsTemp <- as.data.frame(exprsTemp)
    }
    if (!is(exprsTemp,'DelayedMatrix') && argForClassif=="DelayedMatrix"){
        exprsTemp <- DelayedArray::DelayedArray(exprsTemp)
    }
    classResults <- classifier(exprsTemp, clusters, target)
    classResults <- c(unlist(unname(classResults[1])), unlist(unname(classResults[2])))
    typeModified <- classResults[1] != target
    if ( changeType == "not_na")
        typeModified <- (classResults[1] != target) &&
            classResults[1] != "NA"
    if (typeModified && genesModified <= previousGenesModified) {
        tempBestAttackParams <- as.data.frame(t(rowResultsInt))
        colnames(tempBestAttackParams) <- colnames(bestAttackParams)
        bestAttackParams <- tempBestAttackParams
        message( "Better attack with only ", genesModified, " genes modified")
    }
    rowResults <- c(classResults[1], classResults[2],
        genesModified, typeModified, (i + 1), rowResults)
    functionResults <- rbind(rowResults, functionResults)
    list(functionResults=functionResults, bestAttackParams=bestAttackParams)
}

.randWalkNewVector <- function(previousStrComb, modifications,
    bestAttackParams, stepChangeRatio, whileMaxCount, functionResults){
    newWalkParams <- c()
    newWalkStep <- -1
    killCount <- 0
    while (newWalkStep %in% previousStrComb ||
        length(newWalkParams[newWalkParams != (length(modifications) + 1)]) >=
            length(bestAttackParams[bestAttackParams != (length(modifications) + 1)])) {
        newWalkParamsb <- vapply(as.numeric(bestAttackParams), function(param){
            if (param != length(modifications) + 1) {
                if (sample(seq_len(round(1 / stepChangeRatio)), 1) == 1) {
                    return(sample(seq_len(length(modifications) + 1), 1))
                } else {
                    return(param)
                }
            } else {
                if (round(sum(newWalkParams ==
                        (length(modifications) + 1))^(1 / 2) /
                        stepChangeRatio) == 0){
                    return(param)
                } else {
                    if (sample(seq_len(round(sum(newWalkParams ==
                        (length(modifications) + 1))^(1 / 2) /
                        stepChangeRatio)), 1) == 1) {
                        return(sample(seq_len(length(modifications) + 1), 1))
                    } else {
                        return(param)
                    }
                }
            }
        }, numeric(1))
        newWalkParams <- newWalkParamsb
        newWalkStep <- paste(newWalkParams, collapse = " ")
        # If inifite loop kill function and return results
        if (killCount > whileMaxCount) {
            functionResults <- functionResults[order(
                    as.numeric(functionResults$genesModified)), ]
            functionResults <- functionResults[order(
                functionResults$typeModified, decreasing = TRUE), ]
            message("Inifite loop, kill function and return results")
            return(list(FALSE, functionResults))
        }
        killCount <- killCount + 1
    }
    return(list(TRUE, newWalkParams, newWalkStep))
}

.randWalkBeforeWalk <- function(exprs, genes, modifications,
            clusters, target, classifier, firstBatch, argForClassif, argForModif, verbose){
    initObjs <- .initRandWalk(modifications, genes, firstBatch)
    previousStrComb <- initObjs[[1]]
    testsGrid <- initObjs[[2]]
    functionResults <- initObjs[[3]]
    functionResultsInt <- initObjs[[4]]
    results <- initObjs[[5]]
    resultsInt <- initObjs[[6]]
    rwSeed <- .randWalkGetSeed(testsGrid, exprs, genes, modifications,
        clusters, target, classifier, results, resultsInt,
        previousStrComb, argForClassif, argForModif, verbose)
    results <- rwSeed[[1]]
    resultsInt <- rwSeed[[2]]
    previousStrComb <- rwSeed[[3]]
    results$genesModified <- length(genes) - apply(results, 1, function(x) sum(x == "NA"))
    results$typeModified <- results$prediction != target
    results$iteration <- 1
    results <- results[, c("prediction", "odd", "genesModified",
            "typeModified", "iteration", genes)]
    resultsInt$genesModified <- length(genes) - apply( resultsInt, 1,
                        function(x) sum(x == (length(modifications) + 1)))
    resultsInt$typeModified <- resultsInt$prediction != target
    resultsInt$iteration <- 1
    resultsInt <- resultsInt[, c("prediction", "odd", "genesModified",
        "typeModified", "iteration", genes)]
    functionResults <- rbind(results, functionResults)
    functionResults <- functionResults[order(functionResults$genesModified), ]
    functionResults <- functionResults[order(functionResults$typeModified, decreasing = TRUE), ]
    functionResultsInt <- rbind(resultsInt, functionResultsInt)
    functionResultsInt <- functionResultsInt[order(functionResultsInt$genesModified), ]
    functionResultsInt <- functionResultsInt[order(functionResultsInt$typeModified, decreasing = TRUE), ]
    bestAttackParams <- functionResultsInt[1, 6:ncol(functionResultsInt)]
    list(previousStrComb, functionResults, bestAttackParams)
}

.randWalkGetSeed <- function(testsGrid, exprs, genes, modifications, clusters,
    target, classifier, results, resultsInt, previousStrComb, argForClassif, argForModif, verbose){
    i<- 1
    newPreviousStrComb <- c()
    while (i <= nrow(testsGrid)) {
        if (verbose || log2(i) == round(log2(i)) || i == nrow(testsGrid)){
            message("Running first batch to determine walk seed: ", i, " on ", nrow(testsGrid))
        }
        exprsTemp <- exprs
        newPreviousStrComb <- c(newPreviousStrComb, previousStrComb[i])
        rowResults <- c()
        rowResultsInt <- c()
        for (geneInd in seq_along(genes)) {
            modifInd <- testsGrid[i, geneInd]
            if (modifInd != length(modifications) + 1) {
                mod1 <- modifications[[modifInd]][[1]]
                if (length(modifications[[modifInd]]) == 1) {
                    exprsTemp <- advModifications(exprsTemp,
                        genes[geneInd], clusters, target,
                        advMethod = mod1, verbose = verbose)
                } else {
                    mod2 <- modifications[[modifInd]][[2]]
                    exprsTemp <- advModifications(exprsTemp,
                        genes[geneInd], clusters, target,
                        advMethod = mod1,advFixedValue = mod2,
                        advFct = mod2, verbose = verbose)
                }
                rowResults <- c( rowResults,
                    paste(modifications[[modifInd]], collapse = " "))
                rowResultsInt <- c(rowResultsInt, modifInd)
            } else {
                rowResults <- c(rowResults, "NA")
                rowResultsInt <- c(rowResultsInt, modifInd)
            }
        }
        if ( argForClassif == 'SingleCellExperiment'){
            exprsTemp <- SingleCellExperiment(assays = list(counts = t(exprsTemp)))
        }
        if (!is(exprsTemp, 'data.frame') && argForClassif=='data.frame'){
            exprsTemp <- as.data.frame(exprsTemp)
        }
        if (!is(exprsTemp,'DelayedMatrix') && argForClassif=="DelayedMatrix"){
            exprsTemp <- DelayedArray::DelayedArray(exprsTemp)
        }
        classResults <- classifier(exprsTemp, clusters, target)
        classResults <- c(unlist(unname(classResults[1])), unlist(unname(classResults[2])))
        rowResults <- c(rowResults, classResults[1], classResults[2])
        rowResultsInt <- c(unlist(rowResultsInt), classResults[1], classResults[2])
        results <- rbind(results, rowResults)
        resultsInt <- rbind(resultsInt, rowResultsInt)
        colnames(results) <- c(genes, "prediction", "odd")
        colnames(resultsInt) <- c(genes, "prediction", "odd")
        if (classResults[1] == target) {
            i <- i + 1
        } else {
            i <- nrow(testsGrid) + 1
        }
    }
    list(results, resultsInt, newPreviousStrComb)
}

.initRandWalk <- function(modifications, genes, firstBatch){
    previousStrComb <- c(-1)
    if ((length(modifications) + 1)^length(genes) > firstBatch) {
        testsGrid <- data.frame(matrix(ncol = length(genes), nrow = 0))
        while (nrow(testsGrid) < (firstBatch - length(modifications))) {
            randSample <- sample(seq_len(length(modifications) + 1),
                length(genes),
                replace = TRUE
            )
            strRandSample <- paste(randSample, collapse = " ")
            if (!strRandSample %in% previousStrComb) {
                testsGrid <- rbind(testsGrid, randSample)
            }
            previousStrComb <- c(previousStrComb, strRandSample)
        }
        for (i in seq_along(modifications)) {
            testsGrid <- rbind(rep(i, length(genes)), testsGrid)
        }
        testsGrid <- as.matrix(testsGrid)
    } else {
        testsGrid <- gtools::permutations(n = length(modifications) + 1,
            r = length(genes), repeats.allowed = TRUE)
    }
    functionResults <- data.frame(matrix(ncol = length(genes) + 5, nrow = 0))
    colnames(functionResults) <- c("prediction", "odd", "genesModified",
        "typeModified", "iteration", genes)
    functionResultsInt <- data.frame(
        matrix(ncol = length(genes) + 5, nrow = 0)
    )
    colnames(functionResultsInt) <- c( "prediction", "odd",
        "genesModified", "typeModified", "iteration", genes)
    results <- data.frame(matrix(ncol = length(genes) + 2, nrow = 0))
    colnames(results) <- c(genes, "prediction", "odd")
    resultsInt <- data.frame(matrix(ncol = length(genes) + 2, nrow = 0))
    colnames(resultsInt) <- c(genes, "prediction", "odd")
    return (list(previousStrComb, testsGrid, functionResults,
        functionResultsInt, results, resultsInt))
}

#' Find a one gene min change adversarial attack list. A one gene min change
#' adversarial attack  refers to the modification of a single gene
#' within a cluster, leading to a change in its classification. The function
#' returns a list of genes/new classification.
#'
#' @details This function aims to get all genes that when modified individually
#' can lead to a misclassification. You can refer to the
#' 'advModifications' function documentation to more details on how
#' to define a modification.
#' The function is made as a two step parameter search. The first step is to
#' split the genes in 'firstDichot' sets, 100 by default. Then each set is
#' studied by a dichotomic process in a recursive function.
#' The aim of sarting by a high value of sets, instead of starting directly by
#' the dichotomic research is to avoid the following scenario: we modify 5000
#' genes, the modification of one gene conpensates the modification of another.
#' The classification remains unchanged, whereas there is a one gene
#' classification modifying inside the 5000.
#' The dichotomic process runs as follow. The function receives a list of
#' genes, make the modification of the whole list and make the classification.
#' Three scenarios possible:
#' - the classification remains the same as the target cluster. The function
#'  returns, and the dichotomic process continues.
#' - the classification is changed. There is only one gene in the list, the
#'  function returns the gene and the new classification.
#' - the classification is changed. There is more than one gene in the list,
#'  the genes list is split in two, and the dichotomic process continues.
#' 
#' @param exprs DelayedMatrix of numeric RNA expression, cells are rows and genes
#' are columns - or a SingleCellExperiment object, a matrix or a data.frame. By default,
#' these are converted to a data.frame to increase speed performance during modifications.
#' However, this conversion can consume a significant amount of memory, see 'argForModif'
#' argument for options.
#' @param clusters a character vector of the clusters to which the cells belong
#' @param target the name of the cluster to modify
#' @param classifier a classifier in the suitable format.
#' A classifier function should be formated as follow:
#' classifier = function(expr, clusters, target){
#'      # Making the classification
#'      c("cell type", score)
#' }
#' `score` should be numeric between 0 and 1, 1 being the highest confidance
#' into the cell type classification.
#' The matrix `expr` contains RNA expression values, the vector `clusters`
#' consists of the cluster IDs for each cell in `expr`, and `target` is the
#' ID of the cluster for which we want to have a classification.
#' The function returns a vector with the classification result, and a score.
#' @param exclGenes a character vector of genes to exclude from the analysis
#' @param genes a character vector of genes in case you want to limit the
#' attack on a subset of genes
#' @param advMethod the name of the method to use
#' @param advFixedValue the numeric value to use in case of
#' advMethod=`fixed`
#' @param advFct the function to use in case advMethod
#' belongs to the following list: `full_row_fct`, `target_row_fct`,
#' `target_matrix_fct`, `full_matrix_fct`
#' @param firstDichot the initial number of slices before
#' the dichotomic search
#' @param maxSplitSize max size of dichotomic slices
#' @param returnFirstFound set to TRUE to return result when a
#' the first misclassification is found
#' @param changeType `any` consider each misclassification,
#'  `not_na` consider each misclassification but NA.
#' @param argForClassif the type of the first argument to feed to the
#' classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
#' or 'DelayedMatrix'.
#' @param argForModif type of matrix during for the modification, 'data.frame'
#' by default. Can be 'DelayedMatrix', which needs less memory but is slower.
#' @param verbose logical, set to TRUE to activate verbose mode
#' @return a list of genes/new classification tuples
#' @examples
#' library(DelayedArray)
#' 
#' MyClassifier <- function(expr, clusters, target) {
#'    c("B cell", 0.9)
#' }
#' rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
#'      CD8B=c(2,2,3,3)))
#' genes <- c("CD4", "CD8A")
#' clusters_id <- c("B cell","B cell","T cell","T cell")
#' 
#' advSingleGene(rna_expression, clusters_id,
#' "T cell", MyClassifier, advMethod="perc99")
#' @export
advSingleGene <- function(exprs, clusters, target, classifier, exclGenes = c(),
        genes = c(), advMethod = "perc99", advFixedValue = 3,
        advFct = NULL, firstDichot = 100, maxSplitSize = 1,
        returnFirstFound = FALSE, changeType = "any",
        argForClassif = 'data.frame', argForModif = 'data.frame',
        slot=NULL, verbose = FALSE) {
    if (!is(exprs, 'matrix') && !is(exprs,'data.frame') &&
        !is(exprs,'SingleCellExperiment') && !is(exprs,'DelayedMatrix') &&
        !is(exprs,'Seurat')){
        stop("The argument exprs must be a DelayedMatrix, a SingleCellExperiment, a matrix, a data.frame or a Seurat object.")
    }
    if (!is.character(clusters)) {
        stop("The argument clusters must be a vector of character.")
    }
    if (!is.character(target)) {
        stop("The argument target must be character.")
    }
    if (!is.function(classifier)){
        stop("The argument classifier must be a function.")
    }
    if (!is.character(exclGenes) && length(exclGenes)>0) {
        stop("The argument exclGenes must be character or vector of character.")
    }
    if (!is.character(genes) && length(genes)>0) {
        stop("The argument genes must be character or vector of character.")
    }
    if (!is.character(advMethod)) {
        stop("The argument advMethod must be character.")
    }
    if (!is.numeric(firstDichot)) {
        stop("The argument firstDichot must be numeric.")
    }
    if (!is.numeric(maxSplitSize)){
        stop("The argument maxSplitSize must be numeric.")
    }
    if (!is.logical(returnFirstFound)){
        stop("The argument returnFirstFound must be logical.")
    }
    if (!is.character(argForClassif)) {
        stop("The argument argForClassif must be character: 'data.frame' or 'SingleCellExperiment'.")
    }
    if (!is.character(changeType)) {
        stop("The argument changeType must be a vector of character.")
    }
    if (!is.logical(verbose)){
        stop("The argument verbose must be logical.")
    }

    if (is(exprs,'SingleCellExperiment') ){
        exprs <- t(as.matrix(counts(exprs)))
    }
    if (is(exprs,'Seurat') && tolower(argForModif) !="seurat"){
        so_expr <- exprs
        dfScaled <- as.data.frame(exprs@assays$RNA@layers[[slot]])
        colnames(dfScaled) <- rownames(exprs@assays$RNA@cells)
        rownames(dfScaled) <- rownames(exprs@assays$RNA@features)
        dfScaled <- as.data.frame(t(dfScaled))
        exprs <- dfScaled
    }
    if (!is(exprs,'DelayedMatrix') && argForModif=="DelayedMatrix"){
        message("Converting exprs object to a DelayedArray object")
        exprs <- DelayedArray::DelayedArray(exprs)
    }
    if (!is(exprs,'data.frame') && argForModif=="data.frame"){
        message("Converting exprs object to a data.frame object")
        exprs <- as.data.frame(exprs)
    }

    if (tolower(argForClassif) == "seurat"){
        genesIndex <- seq_len(nrow(exprs@assays$RNA@features))
    } else {
        genesIndex <- seq_len(ncol(exprs))
    }

    
    if (length(exclGenes) != 0 && length(genes) == 0) {
        if (tolower(argForClassif) == "seurat"){
            tempGenes <- rownames(exprs@assays$RNA@features)
        } else {
            tempGenes <- colnames(exprs)
        }
        genesToRemove <- unlist(lapply(exclGenes, function(strGene){
            if (!is.na(match(strGene, tempGenes))) {
                return(-match(strGene, tempGenes))
            } else {
                return(NULL)
            }
        }))
        if (length(genesToRemove) != 0)
            genesIndex <- genesIndex[genesToRemove]
    }
    if (length(genes) != 0) {
        if (tolower(argForClassif) == "seurat"){
            tempGenes <- rownames(exprs@assays$RNA@features)
        } else {
            tempGenes <- colnames(exprs)
        }
        exclGenes <- unique(exclGenes)
        genes <- unique(genes)
        genesToKeep <- genes[!genes %in% exclGenes]
        indGenesToKeep <- unlist(lapply(genesToKeep), function(strGene){
            if (!is.na(match(strGene, tempGenes))) {
                return(match(strGene, tempGenes))
            } else {
                return(NULL)
            }
        })
        if (length(indGenesToKeep) != 0)
            genesIndex <- genesIndex[indGenesToKeep]
    }
    if (verbose){message("genesIndex size: ", length(genesIndex))}

    cSplitIndex <- .custSplit(genesIndex, seq_len(firstDichot))
    lastResLength <<- 0
    modObj <- predictWithNewValue(exprs, c("1"),
            clusters, target, classifier,
            advMethod = "none",
            argForClassif = argForClassif, argForModif = argForModif,
            slot=slot, verbose = verbose)
    classifTarget <- unlist(unname(modObj[1]))
    lSplitsResults <- unlist(lapply(seq_len(firstDichot), function(i){
        prevTime <- Sys.time()
        if (lastResLength > 0 && returnFirstFound){
            return(NULL)
        }
        if (verbose || log2(i) == round(log2(i)) || i == firstDichot){
            message(paste0("Split number: ", i, "/", firstDichot))
        }
        if (length(unlist(cSplitIndex[i])) != 0) {
            if (tolower(argForClassif) == "seurat"){
                myGenes <- rownames(exprs@assays$RNA@features)[unlist(cSplitIndex[i])]
            } else {
                myGenes <- colnames(exprs)[unlist(cSplitIndex[i])]
            }

            temp_result <- .predictionDichotMinChange(exprs,
                    myGenes,
                    clusters, target, classifier, advMethod = advMethod,
                    advFixedValue = advFixedValue, advFct = advFct,
                    maxSplitSize = maxSplitSize,
                    returnFirstFound = returnFirstFound,
                    changeType = changeType,
                    argForClassif = argForClassif,
                    argForModif = argForModif,
                    slot=slot,
                    classifTarget = classifTarget,
                    verbose = verbose)
            if ( lastResLength == 0 ){
                lastResLength <<- length(temp_result)
            }
            return(temp_result)
        }
        if (verbose || log2(i) == round(log2(i)) || i == firstDichot){
            message(paste0("Split time: ", Sys.time() - prevTime))
        }
    }), recursive = FALSE)
    if ( is.null(lSplitsResults)){
        NULL
    } else {
        new("advList", values=lSplitsResults)
    }
}

.dichotMinSplit <- function(lResults, cGeneSplitValue, exprs,
            genes, clusters, target, classifier, advMethod, advFixedValue,
            advFct, maxSplitSize, changeType, returnFirstFound, argForClassif, 
            argForModif, slot, classifTarget, verbose){

    if (length(cGeneSplitValue) != 0) {
        if (verbose) {
            message("before predictWithNewValue")
            message(length(cGeneSplitValue))}

        modObj <- predictWithNewValue(exprs, cGeneSplitValue,
            clusters, target, classifier,
            advMethod = advMethod,
            advFixedValue = advFixedValue, advFct = advFct,
            argForClassif = argForClassif, argForModif = argForModif,
            slot=slot, verbose = verbose)
        cellType <- unlist(unname(modObj[1]))
        celltypeScore <- unlist(unname(modObj[2]))
        if (verbose){
            message(cellType)
            message(celltypeScore)
        }
        message("cellType: ", cellType)
        message("classifTarget: ", classifTarget)
        message("target: ", target)
        if (cellType != classifTarget) {
            if (length(cGeneSplitValue) <= maxSplitSize &&
                (changeType == "any" || cellType != "NA" )) {
                if (verbose){
                    message("Store gene:")
                    message(cGeneSplitValue)
                }
                res_key <- paste(cGeneSplitValue, collapse = "__")
                lResults[[res_key]] <- c(cellType, celltypeScore)
            } else {
                if ( length(cGeneSplitValue) > maxSplitSize){
                    if (verbose) {
                        message("Before predictionDichot:")
                        message(length(cGeneSplitValue))
                    }
                    lResults <- .predictionDichotMinChange(exprs,
                        cGeneSplitValue, clusters, target, classifier,
                        lResults = lResults, advMethod = advMethod,
                        advFixedValue = advFixedValue, advFct = advFct,
                        maxSplitSize = maxSplitSize,
                        returnFirstFound=returnFirstFound,
                        changeType = changeType,
                        argForClassif = argForClassif,
                        argForModif=argForModif,
                        slot=slot, classifTarget=classifTarget,
                        verbose = verbose)
                }
            }
        }
    }
    lResults
    
}

.predictionDichotMinChange <- function(exprs, genes, clusters, target,
        classifier, lResults = list(), advMethod = "perc99",
        advFixedValue = 3, advFct = NULL, maxSplitSize = 1,
        returnFirstFound = FALSE, changeType = "any",
        argForClassif, argForModif, slot=NULL,classifTarget=NULL,
        verbose) {
    if ( returnFirstFound && length(lResults) > 0){
        return (lResults)
    }
    cGeneSplit <- .custSplit(unlist(genes), c(1,2))
    if (verbose){
        message(paste0("genes size: ", length(unlist(genes))))
    }
    lResults <- .dichotMinSplit(lResults,
                    unname(unlist(cGeneSplit[1])), exprs,
                    genes, clusters, target, classifier, advMethod,
                    advFixedValue, advFct,
                    maxSplitSize, changeType, returnFirstFound, argForClassif,
                    argForModif, slot, classifTarget, verbose)
    if ( returnFirstFound && length(lResults) > 0){
        return (lResults)
    } 
    lResults <- .dichotMinSplit(lResults,
                    unname(unlist(cGeneSplit[2])), exprs,
                    genes, clusters, target, classifier,
                    advMethod, advFixedValue,
                    advFct, maxSplitSize, changeType,
                    returnFirstFound, argForClassif,
                    argForModif, slot, classifTarget, verbose)
    lResults
}

#' Gives an overview of the susceptibility to single gene
#' attacks, for each cell type, for a given list of modifications.
#'
#' @details Running the advSingleGene function for each cell type
#' to see which ones are more vulerable can take a long time. The
#' aim of the singleGeneOverview function is to make this process faster.
#' It uses a default value of 100 for the 'maxSplitSize' parameter. So,
#' the dichotomic process of the advSingleGene function stops as soon
#' as the fold length is lower than 100. You can have more accurate
#' results with maxSplitSize=1, but it will take longer.
#' This function aims also to run the advSingleGene for several given
#' modifications. You can specify a list of modifications as so - each
#' item of the list should be 1 or 2 length size.
#' The 1 length vector must contain the prerecorded modifications, 'perc1'
#' or 'perc99'.
#' The 2 length vector must have as first item:
#'  - 'fixed', in this case the second item should be the value to be
#'  replaced by.
#'  - 'full_row_fct', 'target_row_fct', 'target_matrix_fct' or
#'  'full_matrix_fct'. In this case the second item should be a function.
#' Let's say we want to analysis the susceptibility to single gene attack
#' for 3 modifications: "perc1", the modification of each value of the
#' cluster by 1000, and a custom modification stored inside a function myFct.
#' Then the 'modification' parameter should be:
#' my_modifications = list(c("perc1"),
#'                         c("fixed", 1000),
#'                         c("full_matrix_fct", myFct))
#' 
#' The function returns a dataframe with the number of genes of the max
#' change attack for each modification in columns, for each cell type in rows.
#' @param exprs DelayedMatrix of numeric RNA expression, cells are rows and genes
#' are columns - or a SingleCellExperiment object, a matrix or a data.frame. By default,
#' these are converted to a data.frame to increase speed performance during modifications.
#' However, this conversion can consume a significant amount of memory, see 'argForModif'
#' argument for options.
#' @param clusters a character vector of the clusters to which the cells belong
#' @param classifier a classifier in the suitable format.
#' A classifier function should be formated as follow:
#' classifier = function(expr, clusters, target){
#'      # Making the classification
#'      c("cell type", score)
#' }
#' `score` should be numeric between 0 and 1, 1 being the highest confidance
#' into the cell type classification.
#' The matrix `expr` contains RNA expression values, the vector `clusters`
#' consists of the cluster IDs for each cell in `expr`, and `target` is the
#' ID of the cluster for which we want to have a classification.
#' The function returns a vector with the classification result, and a score.
#' @param exclGenes a character vector of genes to exclude from the analysis
#' @param genes a character vector of genes in case you want to limit the
#' analysis on a subset of genes
#' @param modifications the list of the modifications to study
#' @param advMethod the name of the method to use
#' @param advFixedValue the numeric value to use in case of
#' advMethod=`fixed`
#' @param advFct the function to use in case advMethod
#' belongs to the following list: `full_row_fct`, `target_row_fct`,
#' `target_matrix_fct`, `full_matrix_fct`
#' @param firstDichot the initial number of slices before
#' the dichotomic search
#' @param maxSplitSize max size of dichotomic slices.
#' @param changeType `any` consider each misclassification,
#'  `not_na` consider each misclassification but NA.
#' @param argForClassif the type of the first argument to feed to the
#' classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
#' or 'DelayedMatrix'.
#' @param argForModif type of matrix during for the modification, 'data.frame'
#' by default. Can be 'DelayedMatrix', which needs less memory but is slower.
#' @param verbose logical, set to TRUE to activate verbose mode
#' @return a DataFrame storing the number of possible single gene
#' attacks each cell type and each modification.
#' @examples
#' library(DelayedArray)
#' 
#' MyClassifier <- function(expr, clusters, target) {
#'    c("T cell", 0.9)
#' }
#' rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
#'      CD8B=c(2,2,3,3)))
#' genes <- c("CD4", "CD8A")
#' clusters_id <- c("B cell","B cell","T cell","T cell")
#'
#' singleGeneOverview(rna_expression, clusters_id,
#' MyClassifier, modifications = list(c("perc1"), c("perc99")))
#' 
#' myModif = function(x, y){
#'    return(sample(1:10,1))
#' }
#' 
#' my_modifications = list(c("perc1"),
#'                         c("fixed", 1000),
#'                         c("full_matrix_fct", myModif))
#' singleGeneOverview(rna_expression, clusters_id,
#'  MyClassifier, modifications = my_modifications)
#' 
#' @export
singleGeneOverview <- function(exprs, clusters, classifier, exclGenes = c(),
            genes = c(), modifications = list(c("perc1"), c("perc99")),
            advMethod = "perc99", advFixedValue = 3, advFct = NULL,
            firstDichot = 100, maxSplitSize = 100, changeType = "any",
            argForClassif = 'data.frame', argForModif = 'data.frame',
            verbose = FALSE) {
    if (!is(exprs, 'matrix') && !is(exprs,'data.frame') &&
        !is(exprs,'SingleCellExperiment') && !is(exprs,'DelayedMatrix')){
        stop("The argument exprs must be a DelayedMatrix, a SingleCellExperiment, a matrix or a data.frame")
    }
    if (!is.character(clusters)) {
        stop("The argument clusters must be a vector of character.")
    }
    if (!is.function(classifier)){
        stop("The argument classifier must be a function.")
    }
    if (!is.character(exclGenes) && length(exclGenes)>0) {
        stop("The argument exclGenes must be character or vector of character.")
    }
    if (!is.character(genes) && length(genes)>0) {
        stop("The argument genes must be character or vector of character.")
    }
    if (!is.list(modifications)) {
        stop("The argument modifications must be a list.")
    }
    if (!is.character(advMethod)) {
        stop("The argument advMethod must be character.")
    }
    if (!is.numeric(firstDichot)){
        stop("The argument firstDichot must be numeric.")
    }
    if (!is.numeric(maxSplitSize)){
        stop("The argument maxSplitSize must be numeric.")
    }
    if (!is.character(changeType)) {
        stop("The argument changeType must be character.")
    }
    if (!is.character(argForClassif)) {
        stop("The argument argForClassif must be character: 'data.frame' or 'SingleCellExperiment'.")
    }
    if (!is.logical(verbose)){
        stop("The argument verbose must be logical.")
    }
    if (is(exprs,'SingleCellExperiment') ){
        exprs <- t(counts(exprs))
    }
    if (!is(exprs,'DelayedMatrix') && argForModif=="DelayedMatrix"){
        message("Converting exprs object to a DelayedArray object")
        exprs <- DelayedArray::DelayedArray(exprs)
    }
    if (!is(exprs,'data.frame') && argForModif=="data.frame"){
        message("Converting exprs object to a data.frame object")
        exprs <- as.data.frame(exprs)
    }

    if (length(modifications) == 0) {
        dfResult <- .minOverArgModifs(exprs, clusters, classifier, exclGenes,
                            genes, advMethod, advFixedValue, advFct,
                            firstDichot, maxSplitSize, changeType, argForClassif, argForModif, verbose)
    } else {
        dfResult <- .minOverListModifs(exprs, clusters, classifier, exclGenes,
                            genes, modifications, firstDichot, maxSplitSize,
                            changeType, argForClassif, argForModif, verbose)
    }
    S4Vectors::DataFrame(dfResult)
}


.minOverListModifs <- function(exprs, clusters, classifier, exclGenes,
                            genes, modifications, firstDichot, maxSplitSize,
                            changeType, argForClassif, argForModif, verbose){
    dfResult <- data.frame(todel = unique(clusters))
    rownames(dfResult) <- unique(clusters);
    dfNames <- unlist(lapply(seq_along(modifications), function(modifInd){
        paste(modifications[[modifInd]], collapse = "_")
    }))
    vecsAttacksLength <- lapply(seq_along(modifications), function(modifInd){
        mod1 <- modifications[[modifInd]][[1]]
        attacksLength <- vapply(unique(clusters), function(cellType){
            if (verbose) {
                message("Running single gene attack on ", cellType, ", with a maxSplitSize of: ", maxSplitSize)
                message("The smaller the maxSplitSize, the more precise the result will be, but it will take longer.")
                message("Modification: ", paste(modifications[[modifInd]], collapse = " "))
            }
            if (length(modifications[[modifInd]]) == 1) {
                minChangeGenes <- advSingleGene(exprs, clusters, cellType,
                    classifier, exclGenes = exclGenes, genes = genes,
                    advMethod = mod1, maxSplitSize = maxSplitSize,
                    firstDichot = firstDichot, changeType = changeType,
                    argForClassif = argForClassif, argForModif=argForModif, verbose = verbose)@values
            } else {
                mod2 <- modifications[[modifInd]][[2]]
                minChangeGenes <- advSingleGene(exprs, clusters, cellType,
                    classifier, exclGenes = exclGenes,
                    genes = genes,
                    advMethod = mod1, advFixedValue = mod2,
                    advFct = mod2,
                    maxSplitSize = maxSplitSize,
                    firstDichot = firstDichot,
                    changeType = changeType,
                    argForClassif = argForClassif,
                    argForModif=argForModif,
                    verbose = verbose)@values
            }
            resultLength <- length(minChangeGenes)
            if (verbose) {
                message("An approximation gives about ", resultLength, " genes can cause a single gene attack on the ",
                    cellType, " cell type for the modification ", paste(modifications[[modifInd]], collapse = " "))
            }
            return(resultLength)
        }, numeric(1))
        return(attacksLength)
    })
    dfResult <- do.call(cbind, list(dfResult, vecsAttacksLength))
    dfResult$todel <- NULL
    colnames(dfResult) <- dfNames
    dfResult
}

.minOverArgModifs <- function(exprs, clusters, classifier, exclGenes,
                    genes, advMethod, advFixedValue, advFct,
                    firstDichot, maxSplitSize, changeType, argForClassif, argForModif, verbose) {
    attacksLength <- vapply(unique(clusters), function(cellType){
        if (verbose) {
            message( "Running single gene attack on ", cellType, ", with a maxSplitSize of: ", maxSplitSize)
            message( "The smaller the maxSplitSize, the more precise the result will be, but it will take longer.")
        }
        minChangeGenes <- advSingleGene(exprs, clusters, cellType,
            classifier,
            exclGenes = exclGenes,
            genes = genes, advMethod = advMethod,
            advFixedValue = advFixedValue,
            advFct = advFct,
            maxSplitSize = maxSplitSize,
            firstDichot = firstDichot,
            changeType = changeType,
            argForClassif = argForClassif,
            argForModif=argForModif,
            verbose = verbose
        )@values
        resultLength <- length(minChangeGenes)
        if (verbose) {
            message( "An approximation gives about ", resultLength, " genes can cause a single gene attack on the ",
                cellType, " cell type")
        }
        return(resultLength)
    }, numeric(1))
    names(attacksLength) <- unique(clusters)
    dfResult <- data.frame(aboutGeneNumber = unlist(attacksLength))
    rownames(dfResult) <- names(attacksLength)
    dfResult
}

## ----installation, eval=FALSE-------------------------------------------------
# ## Install BiocManager is necessary
# if (!require("BiocManager")) {
#     install.packages("BiocManager")
# }
# BiocManager::install('adverSCarial')

## ----load libraries, warning = FALSE, message=FALSE---------------------------
library(adverSCarial)
library(LoomExperiment)

## ----min change attack, message=FALSE, warning = FALSE------------------------
pbmcPath <- system.file("extdata", "pbmc_short.loom", package="adverSCarial")
lfile <- import(pbmcPath, type="SingleCellLoomExperiment")

## ----extract matrix data------------------------------------------------------
matPbmc <- counts(lfile)

## ----visualize1, message=FALSE, warning = FALSE-------------------------------
matPbmc[1:5, 1:5]

## ----load annnots, message=FALSE, warning = FALSE-----------------------------
cellTypes <- rowData(lfile)$cell_type

## ----visualize2, message=FALSE, warning = FALSE-------------------------------
head(cellTypes)

## ----first classif, message=FALSE---------------------------------------------
for ( cell_type in unique(cellTypes)){
    resClassif <- MClassifier(matPbmc, cellTypes, cell_type)
    print(paste0("Expected: ", cell_type, ", predicted: ", resClassif[1]))
}

## ----markers------------------------------------------------------------------
# Known markers for each cell type
markers <- c("IL7R", "CCR7", "CD14", "LYZ", "S100A4", "MS4A1", "CD8A", "FCGR3A", "MS4A7",
              "GNLY", "NKG7", "FCER1A", "CST3", "PPBP")

## ----search min change attack, message=FALSE, warning = FALSE-----------------
genesMinChange <- advSingleGene(matPbmc, cellTypes, "DC",
                        MClassifier, exclGenes = markers, advMethod = "perc99",
                        returnFirstFound = TRUE, changeType = "not_na",
                        firstDichot = 10)

## ----search min change attack bis, warning = FALSE----------------------------
genesMinChange

## ----run min change attack, message=FALSE, warning = FALSE--------------------
matAdver <- advModifications(matPbmc, names(genesMinChange@values)[1], cellTypes, "DC")

## ----verify the min change attack, warning = FALSE, message = FALSE-----------
resClassif <- MClassifier(matAdver, cellTypes, "DC")

## ----verify the min change attack bis, warning = FALSE------------------------
resClassif

## ----search max change attack, warning = FALSE, message=FALSE-----------------
genesMaxChange <- advMaxChange(matPbmc, cellTypes, "Memory CD4 T", MClassifier,
                    exclGenes = markers, advMethod = "perc99")

## ----search max change attack bis, warning = FALSE----------------------------
length(genesMaxChange@values)

## ----run max change attack, message=FALSE, warning=FALSE----------------------
matMaxAdver <- advModifications(matPbmc, genesMaxChange@values, cellTypes, "Memory CD4 T")

## ----verify max change attack, warning = FALSE, message=FALSE-----------------
resClassif <- MClassifier(matMaxAdver, cellTypes, "Memory CD4 T")

## ----verify max change attack bis, warning = FALSE----------------------------
resClassif

## ----CGD1, warning = FALSE----------------------------------------------------
resCGD <- advCGD(as.data.frame(matPbmc), cellTypes,
                "Memory CD4 T", MClassifier, alpha=1, epsilon=1,
                genes=colnames(matPbmc)[ncol(matPbmc):1])

## ----CGD2, warning = FALSE----------------------------------------------------
tail(resCGD$byGeneSummary)

## ----CGD2_bis, warning = FALSE------------------------------------------------
resCGD$oneRowSummary

## ----CGD3, warning = FALSE----------------------------------------------------
modifiedMatrix <- resCGD$expr

## ----CGD4, warning = FALSE----------------------------------------------------
resCGD$modGenes

## ----CGD5, warning = FALSE----------------------------------------------------
MClassifier(modifiedMatrix, cellTypes, "Memory CD4 T")$prediction

## ----session info-------------------------------------------------------------
sessionInfo()

## ----load libraries, warning = FALSE, message=FALSE---------------------------
library(adverSCarial)
library(LoomExperiment)
library(DelayedArray)

## ----min change attack, message=FALSE, warning = FALSE------------------------
pbmcPath <- system.file("extdata", "pbmc_short.loom", package="adverSCarial")
lfile <- import(pbmcPath, type="SingleCellLoomExperiment")

## ----load table---------------------------------------------------------------
matPbmc <- counts(lfile)

## ----visualize1, message=FALSE, warning = FALSE-------------------------------
matPbmc[1:5,1:5]

## ----Load cell type annotations, message=FALSE, warning = FALSE---------------
cellTypes <- rowData(lfile)$cell_type

## ----visualize2, message=FALSE, warning = FALSE-------------------------------
head(cellTypes)

## ----modifications for overview, warning = FALSE------------------------------
modifOutlier <- function(x, y){
    return (max(x)*1000)
}

## ----use the function, warning = FALSE----------------------------------------
modifications <- list(c("perc1"), c("full_row_fct", modifOutlier))

## ----run min change overview, warning = FALSE, message=FALSE------------------
min_change_overview <- singleGeneOverview(matPbmc, cellTypes, MClassifier,
    modifications= modifications, maxSplitSize = 20, firstDichot = 5)

## ----run min change overview bis, warning = FALSE-----------------------------
min_change_overview

## ----run max change overview, warning = FALSE, message=FALSE------------------
max_change_overview <- maxChangeOverview(matPbmc, cellTypes, MClassifier,
    modifications= modifications, maxSplitSize = 20)

## ----run max change overview bis, warning = FALSE-----------------------------
max_change_overview

## ----session info-------------------------------------------------------------
sessionInfo()

#' The function getSignGenes orders the genes by maximizing the
#' significance of the gene to differentiate the clusters and ensures
#' that they represent at most the variations across all possible pairs
#' of clusters.
#' @details The function getSignGenes orders the genes by maximizing the
#' significance of the gene to differentiate the clusters and ensures
#' that they represent at most the variations across all possible pairs
#' of clusters.
#' 
#' The getDistantCouples function is used to generate all possible pairs
#' of clusters and order them in a way that the following pairs are as
#' distant as possible.
#' cell_types = c("B", "CD4 T", "NK", "CD8 T", "DC")
#' getDistantCouples(cell_types)
#'      'B___CD4 T', 'NK___CD8 T', 'B___DC', 'CD4 T___NK',
#'      'CD8 T___DC', 'B___NK', 'CD4 T___CD8 T', 'NK___DC',
#'      'B___CD8 T', 'CD4 T___DC'
#' 
#' @param expr A matrix of gene expression data. Rows are cells and columns
#' are genes.
#' @param clusters a character vector of the clusters to which the cells belong
#' @param method the statistical test to use. Either "wilcox" for the Wilcoxon
#' rank sum test or "ttest" for the t-test. Default is "wilcox".
#' @param verbose logical, set to TRUE to activate verbose mode
#' @examples
#' rna_expression <- data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
#'     CD8B=c(2,2,3,3))
#' genes <- c("CD4", "CD8A")
#' clusters_id <- c("B cell","T cell","T cell","T cell")
#' 
#' getSignGenes(rna_expression, clusters_id, method="wilcox", verbose=TRUE)
#' 
#' @export 
getSignGenes <- function(expr, clusters, method="wilcox", verbose=FALSE){
    combinations <- getDistantCouples(unique(clusters))
    pvalList <- list()
    
    for(strComb in combinations){
        a_a <- unlist(strsplit(strComb, "___"))
        clust1 <- a_a[1]
        clust2 <- a_a[2]
        if (verbose) {message("Cluster ",clust1," vs ",clust2)}
        pvals <- apply(t(expr), 1, function(x){
                c1 <- x[clusters == clust1]
                c1 <- c1[!is.na(c1)]
                c2 <- x[clusters == clust2]
                c2 <- c2[!is.na(c2)]
                if ( length(c1) == 0 || length(c2) == 0){
                    return(1)
                }
                if (length(unique(c1))==1){
                    c1[1] = c1[1] + 0.01
                }
                if (length(unique(c2))==1){
                    c2[1] = c2[1] + 0.01
                }
                if (method=="wilcox"){
                    return(wilcox.test(c1, c2)$p.value)
                }
                if (method=="ttest"){
                    return(t.test(c1, c2)$p.value)
                }    
        })
        dfPvals <- data.frame(gene=colnames(expr), pval=unname(pvals))
        rownames(dfPvals) <- dfPvals$gene
        dfPvals <- dfPvals[order(dfPvals$pval),]
        pvalList[[strComb]] <- dfPvals
    }
    
    dfResults <- as.data.frame(matrix(nrow=0, ncol=2))
    colnames(dfResults) <- c("gene", "pval")
    
    for (i in 1:ncol(expr)){
        for(strCouple in names(pvalList)){
            strGene <- pvalList[[strCouple]][i,"gene"]
            if (!strGene %in% dfResults$gene){
                dfResults <- rbind(dfResults, pvalList[[strCouple]][i,])
            }
        }
    }

    results <- list(results=dfResults, pvalList=pvalList)
    return(results)
}

getDistantCouples <- function(cell_types, verbose=FALSE){
    combinations <- combn(cell_types, 2, simplify = FALSE)
    strCombs <- c()
    for (i in combinations){
        strCombs <- c(strCombs, paste0(i[1], "___", i[2]))
    }
    twoPrevs <- c()
    selCombs <- c()
    intInd <- 1
    killLoop <- FALSE
    intKill <- 0
    while(length(selCombs)<length(combinations)){
        if (intKill == 0 && killLoop){
            selCombs <- c(selCombs, setdiff(strCombs, selCombs))
        }
        intKill <- intKill + 1
        killLoop <- TRUE
        c1 <- combinations[[intInd]][1]
        c2 <- combinations[[intInd]][2]
        myKey <- paste0(c1,"___",c2)

        if ( !c1 %in% twoPrevs && !c2 %in% twoPrevs && !myKey %in% selCombs){
            if (verbose){message(myKey)}
            intKill <- 0
            killLoop <- FALSE
            selCombs <- c(selCombs, myKey)
            twoPrevs <- c(c1, c2)
        }
        intInd <- intInd + 1
        if (intKill > length(combinations) +1){
            intKill <- 0
        }
        if (intInd > length(combinations)){
            intInd <- 1
        }
    }
    return(selCombs)
}

#' adverSCarial class
#' 
#' \code{advChar} is a class used to store the output values
#' of the \code{advMaxChange} function. The result can be a vector
#' of few thousands genes, so a specific *show* method is associated to
#' this class to avoid overflooding the R scripts outputs.
#' @rdname advChar
#' @return A advChar object
#' @examples
#' MyClassifier <- function(expr, clusters, target) {
#'   c("T cell", 0.9)
#' }
#'
#' genes <- paste0("gene_",1:10000)
#' rna_expression <- data.frame(lapply(genes, function(x) numeric(0)))
#' rna_expression <- rbind(rna_expression, rep(1,10000))
#' rna_expression <- rbind(rna_expression, rep(2,10000))
#' colnames(rna_expression) <- genes
#' clusters_id <- c("B cell","T cell")
#'
#' max_change_genes <- advMaxChange(rna_expression, clusters_id,
#' "T cell", MyClassifier, advMethod="perc99")
#'                                    
#' max_change_genes
#' @export
advChar <- methods::setClass("advChar",
        slots = list(
        values = "character"
        ))
setMethod("show", "advChar",
        function(object) {
            if ( length(object@values)<10){
                print(object@values)
            } else {
                print(paste0("Vector with ", length(object@values), " values:"))
                if (is.null(names(object@values))){
                    print(paste0(paste(object@values[1:10], collapse=", "), " ..."))
                } else {
                    print(paste0(paste(paste0(names(object@values[1:5]),": ",object@values[1:5]),
                                        collapse=", "), " ..."))
                }
            }
        })

#' adverSCarial class
#' 
#' \code{advList} is a class used to store the output values
#' of the \code{advSingleGene} function. The result can be a list
#' of few thousands genes:cell_type key/values, so a specific *show*
#' method is associated to this class to avoid overflooding the R
#' scripts outputs.
#' @rdname advList
#' @return A advList object
#' @examples
#' MyClassifier <- function(expr, clusters, target) {
#'    c("B cell", 0.9)
#' }
#' rna_expression <- data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
#'      CD8B=c(2,2,3,3))
#' genes <- c("CD4", "CD8A")
#' clusters_id <- c("B cell","B cell","T cell","T cell")
#' 
#' adv_min_change <- advSingleGene(rna_expression, clusters_id,
#' "T cell", MyClassifier, advMethod="perc99")
#' 
#' adv_min_change
#' 
#' @export
advList <- methods::setClass("advList",
        slots = list(
        values = "list"
        ))
setMethod("show", "advList",
        function(object) {
            if ( length(object@values)<10){
                print(object@values)
            } else {
                print(paste0("Vector with ", length(object@values), " values:"))
                if (is.null(names(object@values))){
                    print(paste0(paste(object@values[1:10], collapse=", "), " ..."))
                } else {
                    print(paste0(paste(paste0(names(object@values[1:5]),": ",object@values[1:5]),
                                        collapse=", "), " ..."))
                }
            }
        })

#' Implementation of the IDG4C algorithm.
#' 
#' @details this function is an implementation of the IDG4C algorithm
#' which permits to generate adversarial attacks on a classifier on a given
#' cluster. The attack is done by modifying the expression of the genes by 
#' gradient descent after having inferred if by the finite difference method.
#' Two parameters alpha and epsilon are used to control the step size of the 
#' modifications.
#' 
#' @param expr a matrix, a data.fram or a Seurat object
#' @param clusters a character vector of the clusters to which the cells belong
#' @param target the name of the cluster to modify
#' @param classifier a classifier in the suitable format.
#' A classifier function should be formated as follow:
#' classifier = function(expr, clusters, target){
#'      # Making the classification
#'      c("cell type", score)
#' }
#' `score` should be numeric between 0 and 1, 1 being the highest confidance
#' into the cell type classification.
#' The matrix `expr` contains RNA expression values, the vector `clusters`
#' consists of the cluster IDs for each cell in `expr`, and `target` is the
#' ID of the cluster for which we want to have a classification.
#' The function returns a vector with the classification result, and a score.
#' @param genes the character vector of genes to study
#' @param exclNewTargets the character vector of cell types to exclude as new target
#' @param newTarget the name of the new cell type cell type target, this will be the
#' new prediction cell type after the attack
#' @param alpha the alpha parameter of the IDG4C algorithm
#' @param epsilon the epsilon parameter of the IDG4C algorithm
#' @param slot the slot to modify in case of Seurat object
#' @param stopAtSwitch logical, set to TRUE to stop the attack when the new target
#' set to FALSE to continue the attack until all genes are tested
#' @param verbose logical, set to TRUE to activate verbose mode
#' @return a list containing the modified expression matrix, the list of modified genes,
#' the summary of the attack by gene, the summary of the attack,
#' the new cell types predictions and the original cell types predictions
#' @examples
#' MyClassifier <- function(expr, clusters, target) {
#'    typePredictions <- as.data.frame(matrix(nrow=nrow(expr), ncol=length(unique(clusters))))
#'    colnames(typePredictions) <- unique(clusters)
#'    typePredictions[unique(clusters)[1]] <- c(1,0,0,0)
#'    typePredictions[unique(clusters)[2]] <- c(0,1,1,1)
#'    rownames(typePredictions) <- 1:4
#'
#'    list(prediction="T cell", odd=1,
#'                      typePredictions=t(typePredictions),
#'                      cellTypes=c("B cell","T cell","T cell","T cell"))
#' }
#' rna_expression <- data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
#'     CD8B=c(2,2,3,3))
#' genes <- c("CD4", "CD8A")
#' clusters_id <- c("B cell","T cell","T cell","T cell")
#' 
#' advCGD(rna_expression, clusters_id, "T cell",
#'   MyClassifier, genes=genes, verbose=TRUE)
#' 
#' 
#' @export
advCGD <- function(expr, clusters, target, classifier, genes=NULL,
                     exclNewTargets=NULL, newTarget=NULL,
                     alpha=0.1, epsilon=0, slot=NULL,
			         stopAtSwitch=TRUE, verbose=FALSE){
    # orExpr <- expr
    predictions <- classifier(expr, clusters, target)
    orTarget <- predictions$prediction
    if (verbose){
      message(orTarget)
      message(dim(t(predictions$typePredictions)))
      message(length(clusters))
      message(sum(clusters == target))
    }
    if (class(expr) == "Seurat"){
        cluster_cell_ids <- rownames(expr@assays$RNA@cells)[clusters == target]
    } else {
        cluster_cell_ids <- names(clusters)[clusters == target]
    }
    
    dfSumUp <- as.data.frame(matrix(nrow=0, ncol=7))
    
    
    cellPredictions <- apply(t(predictions$typePredictions[, clusters == target]), 1, function(x){
            names(x[x == max(x)])[1]
        })
    oriCellPredictions <- cellPredictions
    #display(table(unlist(cellPredictions)))
#     display(names(cellPredictions[cellPredictions=='Myeloid Dendritic cells']))

    # Define the cell type after the attack
    if (is.null(newTarget)){
        orTable <- table(cellPredictions)
        orTable <- orTable[order(orTable, decreasing=T)]
        if (length(names(orTable)) == 1 ){
            # only on cell type predicted, so take the second cell type with the higher odd
            clustTypeOdds <- apply(predictions$typePredictions[, clusters == target], 1, function(x){
                mean(x)
            })
            clustTypeOdds <- clustTypeOdds[order(clustTypeOdds, decreasing = TRUE)]
            if (names(clustTypeOdds)[1] == orTarget){
                newTarget <- names(clustTypeOdds)[2]
            } else {
                newTarget <- names(clustTypeOdds)[1]
            }
        } else {
            # take second most present predicted cell type
            newTarget <- setdiff(names(orTable), c(exclNewTargets, orTarget))[1]
        }
    }
    rightCells <- names(cellPredictions[cellPredictions==newTarget])
    orCells <- names(cellPredictions[cellPredictions==orTarget])
    nOrTarg <- length(orCells)
    c4summaryStart <- c(nOrTarg/sum(clusters==target),
                    length(rightCells)/sum(clusters==target))
    prevMeanOrCell <- mean(unlist(predictions$typePredictions[orTarget,orCells]))
    if ( length(rightCells) == 0){
        prevMeanRightCell <- 0
    } else {
        prevMeanRightCell <- mean(unlist(predictions$typePredictions[newTarget,rightCells]))
    }
    deltMeanRightCell <- 0
    deltMeanOrCell <- 0
    
    if(verbose){message("New cluster target: ",newTarget)}

    
    if (is.null(genes) && class(expr) != "Seurat"){
        genes <- colnames(expr)[sample(1:ncol(expr), ncol(expr))]
    }
    if ( is.null(genes) && class(expr) == "Seurat"){
        genes <- expr@assays$RNA@features[sample(1:length(expr@assays$RNA@features), length(expr@assays$RNA@features))]
   }

    modExpr <- expr
    tempExpr <- expr
    modifiedGenes <- c()
    geneCount <- 1
    geneInDS <- TRUE
    dfSumUp <- rbind(dfSumUp, c(0, "#", 0, nOrTarg/sum(clusters==target),
                              length(rightCells)/sum(clusters==target),
                               prevMeanOrCell, prevMeanRightCell,
                               mean(unlist(predictions$typePredictions[orTarget,cluster_cell_ids])),
                              mean(unlist(predictions$typePredictions[newTarget,cluster_cell_ids]))))
    for(myGene in genes){
        if(verbose){
            message(myGene, " ", geneCount)
            message("Number of original annot ", orTarget," : ", nOrTarg)
            message(" mean ",  prevMeanOrCell, " delt ", deltMeanOrCell)
            message("Number of ", newTarget, " : ", length(rightCells))
            message(" mean ", prevMeanRightCell, " delt ", deltMeanRightCell)
            
        }
        
        if (class(expr) == "Seurat"){
            dfScaled <- as.data.frame(modExpr@assays$RNA@layers[[slot]])
            colnames(dfScaled) <- rownames(modExpr@assays$RNA@cells)
            rownames(dfScaled) <- rownames(modExpr@assays$RNA@features)
            dfScaled <- as.data.frame(t(dfScaled))
            # modify only cells out of target cells
            cells2modify <- rownames(dfScaled)[clusters == target]
            cells2modify <- setdiff(cells2modify, rightCells)
            if ( !myGene %in% colnames(dfScaled)){
                myGene <- sub( "\\.","-",myGene)
                if (!myGene %in% colnames(dfScaled)){
                    geneInDS <- FALSE
                }
            }
            if (geneInDS){
                dfScaled[cells2modify, myGene][dfScaled[cells2modify, myGene]>0] <- dfScaled[cells2modify, myGene][dfScaled[cells2modify, myGene]>0]*(1+alpha) + epsilon
                dfScaled[cells2modify, myGene][dfScaled[cells2modify, myGene]<= 0] <- dfScaled[cells2modify, myGene][dfScaled[cells2modify, myGene] <= 0]*(1-alpha) + epsilon
                modExpr@assays$RNA@layers[[slot]] <- t(dfScaled)
            }
        } else {
            # modify only cells out of target cells
            cells2modify <- rownames(modExpr)[clusters == target]
            cells2modify <- setdiff(cells2modify, rightCells)
            # modExpr[cells2modify, myGene] <- modExpr[cells2modify, myGene]*(1+alpha) + epsilon    
            modExpr[cells2modify, myGene][modExpr[cells2modify, myGene]>0] <- modExpr[cells2modify, myGene][modExpr[cells2modify, myGene]>0]*(1+alpha) + epsilon
            modExpr[cells2modify, myGene][modExpr[cells2modify, myGene]<= 0] <- modExpr[cells2modify, myGene][modExpr[cells2modify, myGene] <= 0]*(1-alpha) + epsilon
        }
        
        newPredictions <- classifier(modExpr, clusters, target)
        newCellPredictions <- apply(t(newPredictions$typePredictions[, clusters == target]), 1, function(x){
            names(x[x == max(x)])[1]
        })
        
        dfnp <- as.data.frame(t(newPredictions$typePredictions[,colnames(predictions$typePredictions[, cells2modify])]))
        if (nrow(dfnp) == 0){
            message("break")
            break
        }
        dfp <- t(predictions$typePredictions[, cells2modify])
        allCols <- unique(c(colnames(dfnp), colnames(dfp)))
        dfnp[,setdiff(allCols, colnames(dfnp))] <- 0
        dfp[,setdiff(allCols, colnames(dfp))] <- 0
        dfnp <- dfnp[, order(colnames(dfnp))]
        dfp <- dfp[,colnames(dfnp)]
        
#         display(head(dfnp))
#         display(head(dfp))
#         display(all(rownames(dfnp) == rownames(dfp)))
#         display(all(colnames(dfnp) == colnames(dfp)))
        
        # Make sure colnames are the same, same length
        colnames(dfnp) <- paste0("new_",colnames(dfnp))
        colnames(dfp) <-paste0("or_",colnames(dfp))
        df <- cbind(dfnp, dfp)
        scoreDelts <- t(apply(df, 1, function(x){
            newVals <- x[grep("^new_", names(x), value=TRUE)]
            orVals <- x[grep("^or_", names(x), value=TRUE)]
            newVals - orVals
        }))
        

#         display(head(df2disp[,c("newCellPredictions", "new_Myeloid Dendritic cells", colnames(scoreDelts))]))

#         cellNewPredictions <- apply(t(newPredictions$typePredictions[, clusters == target]), 1, function(x){
#             names(x[x == max(x)])[1]
#         })
        # If attack succeeded
            # try attack on tempExpr
                # return result with summary of what has changed
        # Which cells benefit from modif
        scoreDelts <- as.data.frame(scoreDelts)
        scoreDelts$cell_id <- rownames(scoreDelts)
        scoreDelts$prevCellPredictions <- cellPredictions[cells2modify]

        # Which cells benefit from modif
        cbfm <- unlist(apply(scoreDelts[scoreDelts$prevCellPredictions != newTarget,], 1, function(x){
            if (as.numeric(unname(x[paste0("new_",newTarget)])) > as.numeric(unname(x[paste0("new_",orTarget)]))) {
                return(x["cell_id"])
            }
        }))
        # Which cells benefit from opposit modif
        cbfom <- unlist(apply(scoreDelts[scoreDelts$prevCellPredictions != newTarget,], 1, function(x){
            if (as.numeric(unname(x[paste0("new_",newTarget)])) < as.numeric(unname(x[paste0("new_",orTarget)]))) {
                return(x["cell_id"])
            }
        }))
        
        if (verbose){
            message("Number of modified cells ", (length(cbfm) + length(cbfom)))
        }
        modExpr <- tempExpr
        
        if ((length(cbfm) + length(cbfom)) > 0){

            modifiedGenes <- c(modifiedGenes, myGene)
            if (class(expr) == "Seurat"){
                dfScaled <- as.data.frame(modExpr@assays$RNA@layers[[slot]])
                colnames(dfScaled) <- rownames(modExpr@assays$RNA@cells)
                rownames(dfScaled) <- rownames(modExpr@assays$RNA@features)
                dfScaled <- as.data.frame(t(dfScaled))
                dfScaled[cbfm, myGene][dfScaled[cbfm, myGene]>0] <- dfScaled[cbfm, myGene][dfScaled[cbfm, myGene]>0]*(1+alpha) + epsilon
                dfScaled[cbfm, myGene][dfScaled[cbfm, myGene]<= 0] <- dfScaled[cbfm, myGene][dfScaled[cbfm, myGene] <= 0]*(1-alpha) + epsilon
#                 dfScaled[cbfm, myGene] <- dfScaled[cbfm, myGene]*(1+alpha) + epsilon
                dfScaled[cbfom, myGene][dfScaled[cbfom, myGene]>0] <- dfScaled[cbfom, myGene][dfScaled[cbfom, myGene]>0]*(1-alpha) - epsilon
                dfScaled[cbfom, myGene][dfScaled[cbfom, myGene]<= 0] <- dfScaled[cbfom, myGene][dfScaled[cbfom, myGene] <= 0]*(1+alpha) - epsilon
#                 dfScaled[cbfom, myGene] <- dfScaled[cbfom, myGene]*(1-alpha) - epsilon
                modExpr@assays$RNA@layers[[slot]] <- t(dfScaled)
            } else {
                modExpr[cbfm, myGene] <- modExpr[cbfm, myGene]*(1+alpha) + epsilon
                modExpr[cbfom, myGene] <- modExpr[cbfom, myGene]*(1-alpha) - epsilon
            }
            
            tempExpr <- modExpr

            tempPredictions <- classifier(modExpr, clusters, target)
            predictions <- tempPredictions
            cellPredictions <- apply(t(tempPredictions$typePredictions[, clusters == target]),
                                     1, function(x){
                names(x[x == max(x)])[1]
            })

            orTable <- table(cellPredictions)
            orTable <- orTable[order(orTable, decreasing=T)]
            if (length(names(orTable)) == 1 ){
                # only on cell type predicted, so take the second cell type with the higher odd
                clustTypeOdds <- apply(predictions$typePredictions[, clusters == target], 1, function(x){
                    mean(x)
                })
                clustTypeOdds <- clustTypeOdds[order(clustTypeOdds, decreasing = TRUE)]
                if (names(clustTypeOdds)[1] == orTarget){
                    tempNewTarget <- names(clustTypeOdds)[2]
                } else {
                    tempNewTarget <- names(clustTypeOdds)[1]
                }
            } else {
                # take second most present predicted cell type
                tempNewTarget <- setdiff(names(orTable), c(exclNewTargets, orTarget))[1]
            }
            if(tempNewTarget != newTarget && verbose){
                message("New cluster target: ",tempNewTarget)
            }
            newTarget <- tempNewTarget
            rightCells <- names(cellPredictions[cellPredictions==newTarget])
            if (length(rightCells) == 0){
                tempPrevMeanRightCell <- 0
            } else {
                tempPrevMeanRightCell <- mean(unlist(tempPredictions$typePredictions[newTarget,rightCells]))
            }
            deltMeanRightCell <- tempPrevMeanRightCell - prevMeanRightCell
            prevMeanRightCell <- tempPrevMeanRightCell
            
            orCells <- names(cellPredictions[cellPredictions==orTarget])
            nOrTarg <- length(orCells)
            tempPrevMeanOrCell <- mean(unlist(tempPredictions$typePredictions[orTarget,orCells]))
            #display("if if if")
            #display(tempPrevMeanOrCell)
            #display(prevMeanOrCell)
            deltMeanOrCell <- tempPrevMeanOrCell - prevMeanOrCell
            prevMeanOrCell <- tempPrevMeanOrCell
            
            if (length(rightCells)>nOrTarg && stopAtSwitch){
                #display(1)
                #display(table(cellPredictions)[order(table(cellPredictions), decreasing = T)])
                
                dfSumUp <- rbind(dfSumUp, c(geneCount, myGene,
                    (length(cbfm) + length(cbfom)),
                    nOrTarg/sum(clusters==target),
                    length(rightCells)/sum(clusters==target),
                    prevMeanOrCell, prevMeanRightCell,
                    mean(unlist(tempPredictions$typePredictions[orTarget,cluster_cell_ids])),
                    mean(unlist(tempPredictions$typePredictions[newTarget,cluster_cell_ids]))))
                colnames(dfSumUp) <- c("order","gene", "nbModifiedCells",
                   "oriCell_pourc", "targCell_pourc",
                   "oriCell_odd", "targCell_odd",
                   "wholeClust_oriOdd", "wholeClust_targOdd")
                cSummary <- c(orTarget,sum(clusters==target), newTarget, alpha, epsilon, 0,
                    length(modifiedGenes), geneCount,
                    c4summaryStart,
                    nOrTarg/sum(clusters==target),
                    length(rightCells)/sum(clusters==target),
                    dfSumUp[1, "wholeClust_oriOdd"],
                    dfSumUp[1, "wholeClust_targOdd"],
                    dfSumUp[nrow(dfSumUp), "wholeClust_oriOdd"],
                    dfSumUp[nrow(dfSumUp), "wholeClust_targOdd"]
                    )
                return(list(expr=modExpr, modGenes=modifiedGenes,
                    byGeneSummary=dfSumUp, oneRowSummary=cSummary,
                    modCellTypes=cellPredictions,
                    oriCellTypes=oriCellPredictions))
            }
            dfSumUp <- rbind(dfSumUp, c(geneCount, myGene,
                (length(cbfm) + length(cbfom)),
                nOrTarg/sum(clusters==target),
                length(rightCells)/sum(clusters==target),
                prevMeanOrCell, prevMeanRightCell,
                mean(unlist(tempPredictions$typePredictions[orTarget,cluster_cell_ids])),
                mean(unlist(tempPredictions$typePredictions[newTarget,cluster_cell_ids]))))
        } else {
            deltMeanOrCell <- 0
            deltMeanRightCell <- 0
            colnames(dfSumUp) <- c("order", "gene", "nbModifiedCells",
                   "oriCell_pourc", "targCell_pourc",
                   "oriCell_odd", "targCell_odd",
                   "wholeClust_oriOdd", "wholeClust_targOdd")
            dfSumUp <- rbind(dfSumUp, c(geneCount, myGene, (length(cbfm) + length(cbfom)),
                    nOrTarg/sum(clusters==target),
                    length(rightCells)/sum(clusters==target),
                    prevMeanOrCell, prevMeanRightCell,
                    dfSumUp[nrow(dfSumUp), "wholeClust_oriOdd"],
                    dfSumUp[nrow(dfSumUp), "wholeClust_targOdd"])
                    )
        }

        geneCount <- geneCount + 1
    }
    
    #display(2)
    tempPredictions <- classifier(modExpr, clusters, target)
    cellPredictions <- apply(t(tempPredictions$typePredictions[, clusters == target]),
                                 1, function(x){
            names(x[x == max(x)])[1]
        })
    #display(table(cellPredictions)[order(table(cellPredictions), decreasing = T)])
    colnames(dfSumUp) <- c("order","gene", "nbModifiedCells",
                       "oriCell_pourc", "targCell_pourc",
                       "oriCell_odd", "targCell_odd",
                   "wholeClust_oriOdd", "wholeClust_targOdd")
    
    cSummary <- c(orTarget, sum(clusters==target), newTarget, alpha, epsilon, 0,
                length(modifiedGenes), length(genes),
                 c4summaryStart, nOrTarg/sum(clusters==target),
                 length(rightCells)/sum(clusters==target),
                 # mean odd of ori cell type in the cluster at the beginning and at the end
                 dfSumUp[1, "wholeClust_oriOdd"],
                 dfSumUp[1, "wholeClust_targOdd"],
                 dfSumUp[nrow(dfSumUp), "wholeClust_oriOdd"],
                 dfSumUp[nrow(dfSumUp), "wholeClust_targOdd"]
                 )
                  
    return(list(expr=modExpr, modGenes=modifiedGenes,
                byGeneSummary=dfSumUp, oneRowSummary=cSummary,
                modCellTypes=cellPredictions,
                oriCellTypes=oriCellPredictions))
}

#' Grid search of min change adversarial attack. Tries each
#' combination on a cluster, given a list of genes and a list of modifications.
#'
#' @details This function aims to find the shortest combination of genes
#' allowing to make a min change attack. It will test every possible combination
#' for a given gene list. This function can take a long time to run, and we
#' recommand to use the random walk search advRandWalkMinChange function instead
#' for lists above 10 genes.
#' 
#' You can specify a list of modifications as so, each item of the list should be 1 or 2 length size.
#' The 1 length vector must contain the prerecorded modifications, 'perc1' or 'perc99'.
#' The 2 length vector must have as first item:
#'  - 'fixed', in this case the second item should be the value to be
#'  replaced by.
#'  - 'full_row_fct', 'target_row_fct', 'target_matrix_fct' or
#'  'full_matrix_fct'. In this case the second item should be a function.
#' Let's say we want to analysis the susceptibility to min change attack
#' for 3 modifications: "perc1", the modification of each value of the
#' cluster by 1000, and a custom modification stored inside a function myFct.
#' Then the 'modification' parameter should be:
#' my_modifications = list(c("perc1"),
#'                         c("fixed", 1000),
#'                         c("full_matrix_fct", myFct))
#' @param exprs DelayedMatrix of numeric RNA expression, cells are rows and genes
#' are columns - or a SingleCellExperiment object, a matrix or a data.frame. By
#' default matrix and data.frame are converted to DelayedMatrix for memory
#' performance, see 'argForModif' argument for options.
#' @param clusters a character vector of the clusters to which the cells belong
#' @param target the name of the cluster to modify
#' @param classifier a classifier in the suitable format.
#' A classifier function should be formated as follow:
#' classifier = function(expr, clusters, target){
#'      # Making the classification
#'      c("cell type", score)
#' }
#' `score` should be numeric between 0 and 1, 1 being the highest confidance
#' into the cell type classification.
#' The matrix `expr` contains RNA expression values, the vector `clusters`
#' consists of the cluster IDs for each cell in `expr`, and `target` is the
#' ID of the cluster for which we want to have a classification.
#' The function returns a vector with the classification result, and a score.
#' @param genes the character vector of genes to study
#' @param modifications the list of the modifications to study
#' @param returnFirstFound set to TRUE to return result when a
#' the first misclassification is found
#' @param argForClassif the type of the first argument to feed to the
#' classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
#' or 'DelayedMatrix'.
#' @param argForModif type of matrix during for the modification, 'data.frame'
#' by default. Can be 'DelayedMatrix', which needs less memory but is slower.
#' @param verbose logical, set to TRUE to activate verbose mode
#' @param iamsure logical, prevents from expansive calculations
#' when `genes` list is too long, set to `TRUE` to run anyway.
#' @return DataFrame results of the classification of all the grid combinations
#' @examples
#' library(DelayedArray)
#' 
#' MyClassifier <- function(expr, clusters, target) {
#'    c("T cell", 0.9)
#' }
#' rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
#'      CD8B=c(2,2,3,3)))
#' genes <- c("CD4", "CD8A")
#' clusters_id <- c("B cell","B cell","T cell","T cell")
#' 
#' advGridMinChange(rna_expression, clusters_id, "T cell",
#'  MyClassifier, genes=genes,
#'  modifications = list(c("perc1"), c("perc99")))
#' 
#' myModif = function(x, y){
#'    return(sample(1:10,1))
#' }
#' 
#' my_modifications = list(c("perc1"),
#'                         c("fixed", 1000),
#'                         c("full_matrix_fct", myModif))
#' advGridMinChange(rna_expression, clusters_id, "T cell",
#'  MyClassifier, genes=genes, modifications = my_modifications)
#' 
#' @export
advGridMinChange <- function(exprs, clusters, target, classifier,
                genes, modifications = list(c("perc1"), c("perc99")),
                returnFirstFound = FALSE, argForClassif = 'data.frame',
                argForModif = 'data.frame',
                verbose = FALSE, iamsure = FALSE) {
    if (!is(exprs, 'matrix') && !is(exprs,'data.frame') &&
        !is(exprs,'SingleCellExperiment') && !is(exprs,'DelayedMatrix')){
        stop("The argument exprs must be a DelayedMatrix, a SingleCellExperiment, a matrix or a data.frame")
    }
    if (!is.character(clusters)) {
        stop("The argument clusters must be a vector of character.")
    }
    if (!is.character(target)) {
        stop("The argument target must be character.")
    }
    if (!is.function(classifier)){
        stop("The argument classifier must be a function.")
    }
    if (!is.character(genes)) {
        stop("The argument genes must be character or vector of character.")
    }
    if (!is.list(modifications)) {
        stop("The argument modifications must be a list.")
    }
    if (!is.logical(returnFirstFound)){
        stop("The argument returnFirstFound must be logical.")
    }
    if (!is.character(argForClassif)) {
        stop("The argument argForClassif must be character: 'data.frame' or 'SingleCellExperiment'.")
    }
    if (!is.logical(verbose)){
        stop("The argument verbose must be logical.")
    }
    if (!is.logical(iamsure)){
        stop("The argument iamsure must be logical.")
    }

    if (is(exprs,'SingleCellExperiment') ){
        exprs <- t(as.matrix(counts(exprs)))
    }
    if (!is(exprs,'DelayedMatrix') && argForModif=="DelayedMatrix"){
        message("Converting exprs object to a DelayedArray object")
        exprs <- DelayedArray::DelayedArray(exprs)
    }
    if (!is(exprs,'data.frame') && argForModif=="data.frame"){
        message("Converting exprs object to a data.frame object")
        exprs <- as.data.frame(exprs)
    }

    functionResults <- data.frame(matrix(ncol = length(genes) + 5, nrow = 0))
    colnames(functionResults) <- c("prediction", "odd", "genesModified",
        "typeModified", "iteration", genes)
    if (.gridWarning(modifications, genes, iamsure))
        return(functionResults)
    testsGrid <- gtools::permutations(n = length(modifications) + 1,
        r = length(genes), repeats.allowed = TRUE)
    results <- data.frame(matrix(ncol = length(genes) + 2, nrow = 0))
    colnames(results) <- c(genes, "prediction", "odd")
    i <- 1
    while (i <= nrow(testsGrid)) {
        if (verbose || log2(i) == round(log2(i)) || i == nrow(testsGrid)){
            message("Running combination: ", i, " on ", nrow(testsGrid))
        }
        exprsTemp <- exprs
        rowResults <- c()
        for (geneInd in seq_along(genes)) {
            modifInd <- testsGrid[i, geneInd]
            if (modifInd != length(modifications) + 1) {
                mod1 <- modifications[[modifInd]][[1]]
                if (length(modifications[[modifInd]]) == 1) {
                    exprsTemp <- advModifications(exprsTemp, genes[geneInd],
                        clusters, target, advMethod = mod1, verbose = verbose,
                        argForModif=argForModif)
                } else {                    
                    mod2 <- modifications[[modifInd]][[2]]
                    exprsTemp <- advModifications(exprsTemp, genes[geneInd],
                        clusters, target, advMethod = mod1, advFct = mod2,
                        advFixedValue = mod2, verbose = verbose,
                        argForModif=argForModif)}
                    rowResults <- c(rowResults,
                        paste(modifications[[modifInd]], collapse = " "))
            } else { rowResults <- c(rowResults, "NA")}
        }
        if ( argForClassif == 'SingleCellExperiment'){
            exprsTemp <- SingleCellExperiment(assays = list(counts = t(exprsTemp)))
        }
        if (!is(exprsTemp, 'data.frame') && argForClassif=='data.frame'){
            exprsTemp <- as.data.frame(exprsTemp)
        }
        if (!is(exprsTemp,'DelayedMatrix') && argForClassif=="DelayedMatrix"){
            exprsTemp <- DelayedArray::DelayedArray(exprsTemp)
        }
        classResults <- classifier(exprsTemp, clusters, target)
        classResults <- c(unlist(unname(classResults[1])),
                            unlist(unname(classResults[2])))
        rowResults <- c(rowResults, classResults[1], classResults[2])
        results <- rbind(results, rowResults)
        colnames(results) <- c(genes, "prediction", "odd")
        ifelse(classResults[1] == target, i <- i + 1, i <- nrow(testsGrid)+1)
    }
    results$genesModified <- length(genes) - apply(results, 1, function(x) sum(x == "NA"))
    results$typeModified <- results$prediction != target
    results <- results[, c( "prediction", "odd",
            "genesModified", "typeModified", genes)]
    functionResults <- rbind(results, functionResults)
    functionResults <-
        functionResults[order(as.numeric(functionResults$genesModified)), ]
    functionResults <- functionResults[order(functionResults$typeModified,
            decreasing = TRUE), ]
    S4Vectors::DataFrame(functionResults)
}

.gridWarning <- function(modifications, genes, iamsure){
    if ((length(modifications) + 1)^length(genes) > 100000 & !iamsure) {
        message("Exit because of too many combinations to test: ", (length(modifications) + 1)^length(genes))
        message("This will probably make your computer freeze.")
        message("You should lower the number of genes and/or of modifications. For example 5 genes and 2 modifications",
            " gives 243 combinations to test")
        message("You can use the iamsure=TRUE option to run the function anyway.")
        return(TRUE)
    } else {
        return (FALSE)
    }
}

#' Returns the RNA expression matrix from a SingleCellExperiment
#' with unique hgnc gene names in columns
#'
#' @details This function retrieves from a SingleCellExperiment
#' object the raw RNA expression value corresponding to the hgnc
#' genes. The resulting matrix can then be used with adverSCarial
#' packages.
#' @param sce SingleCellExperiment object to convert
#' @return the RNA expression matrix from a SingleCellExperiment
#' with unique hgnc gene names in columns
#' @examples
#' library(TENxPBMCData)
#'
#' pbmc <- TENxPBMCData(dataset = "pbmc3k")
#' mat_rna <- matrixFromSCE(pbmc)
#'
#' @export
matrixFromSCE <- function(sce) {
    if ( !is(sce,'SingleCellExperiment')){
        stop("The argument sce must be a SingleCellExperiment.")
    }
    # symbols <- sce@rowRanges@elementMetadata@listData$Symbol_TENx
    symbols <- sce@rowRanges@elementMetadata@listData$Symbol
    ind2keep <- setdiff(match(unique(symbols), symbols), which(is.na(symbols)))

    df <- data.frame(sce@assays@data$counts)
    df <- df[ind2keep, ]
    rownames(df) <- sce@ rowRanges@ elementMetadata@listData$Symbol[ind2keep]
    colnames(df) <- colData(sce)$Barcode
    t(df)
}

#' Example cell type classifier for pbmc clustered datasets.
#'
#' @details This classifier aims at testing the adverSCarial
#' package of real pbmc data. It is a simple marker based
#' classifier. It looks at the average value of a few genes
#' inside a cluster, and returns the associated cell type.
#' Markers where found by differential expressions.
#' @param exprs DelayedMatrix of numeric RNA expression, cells are rows and genes
#' are columns - or a SingleCellExperiment object, a matrix or a data.frame.
#' @param clusters vector of clusters to which each cell belongs
#' @param target name of the cell cluster to classify
#' @return a vector with the classification, and the odd
#' @examples
#' library(TENxPBMCData)
#'
#' pbmc <- TENxPBMCData(dataset = "pbmc3k")
#' mat_rna <- matrixFromSCE(pbmc)
#' cell_types <- system.file("extdata",
#'     "pbmc3k_cell_types.tsv",
#'     package = "adverSCarial"
#' )
#' cell_types <- read.table(cell_types, sep = "\t")$cell_type
#'
#' MClassifier(mat_rna, cell_types, "DC")
#'
#' @export
MClassifier <- function(exprs, clusters, target) {
    if (!is(exprs, 'matrix') && !is(exprs,'data.frame') &&
        !is(exprs,'SingleCellExperiment') && !is(exprs,'DelayedMatrix')){
        stop("The argument exprs must be a DelayedMatrix, a SingleCellExperiment, a matrix or a data.frame")
    }
    if (is(exprs,'SingleCellExperiment') ){
        exprs <- t(counts(exprs))
    }

    res_cell_type <- "UNDETERMINED"
    if (!is(target,"character")) {
        stop("The argument target must be character.")
    }
    if (!is(clusters,"character")) {
        stop("The argument clusters must be a vector of character.")
    }
    if (mean(exprs[clusters == target, "LTB"]) > 7) {
        res_cell_type <- "Memory CD4 T"
    }
    if (mean(exprs[clusters == target, "CD79A"]) > 2) {
        res_cell_type <- "B"
    }
    if (mean(exprs[clusters == target, "S100A9"]) > 10) {
        res_cell_type <- "CD14+ Mono"
    }
    if (mean(exprs[clusters == target, "GZMB"]) > 5) {
        res_cell_type <- "NK"
    }
    if (mean(exprs[clusters == target, "GZMK"]) > 2) {
        res_cell_type <- "CD8 T"
    }
    if (mean(exprs[clusters == target, "LDHB"]) > 3 &&
        mean(exprs[clusters == target, "LDHB"]) < 3.3) {
        res_cell_type <- "Naive CD4 T"
    }
    if (mean(exprs[clusters == target, "CCR7"]) > 0.5) {
        res_cell_type <- "Naive CD4 T"
    }

    if (mean(exprs[clusters == target,"LST1"]) > 10) {
        res_cell_type <- "FCGR3A+ Mono"
    }
    if (mean(exprs[clusters == target, "CD74"]) > 50) {
        res_cell_type <- "DC"
    }
    if (mean(exprs[clusters == target, "PF4"]) > 10) {
        res_cell_type <- "Platelet"
    }

    df_tp <- as.data.frame(matrix(0,
        nrow=nrow(exprs), ncol=length(unique(clusters))))
    colnames(df_tp) <- unique(clusters)
    rownames(df_tp) <- rownames(exprs)
    df_tp[,res_cell_type] <- 1
    df_tp <- t(df_tp)

    resClass <- list(prediction=res_cell_type, odd=1,
                      typePredictions=df_tp,
                      cellTypes=rep(res_cell_type, sum(clusters == target)))
    return(resClass)
}

#' Find a max change adversarial attack. It finds the longer
#' list of genes you can modify on a cluster without changing its
#' classification.
#' 
#' @details This function aims to get the largest part of the genes
#' that can be modified without altering the classification,
#' considering a given modification. You can refer to the
#' 'advModifications' function documentation to more details on how
#' to define a modification.
#' The search is made by a dichotomic process, on a reccursive function.
#' At each iteration the function splits the genes in two groups. It
#' proceeds to the modification of the RNA gene value of the first group,
#' makes its classification. Then three possible scenarios:
#' - the classification is the same as the target cluster. We concat
#'  the genes list to the previous one, make the classification, and it still
#'  gives same classification. Then we return the genes list.
#' - the classification is the same as the target cluster. We concat
#'  the genes list to the previous one, make the classification, and it
#'  gives a different classification. This happens often, you can modify
#'  the gene A with a classification of T cell, or modify the gene B with a
#'  classification of T cell, but modifying A and B returns another
#'  classification. In this case we split the genes list in two and try again.
#' - the classification is not the same as the target cluster. In this case we
#'  split the genes list in two and try again.
#' The iteration process stops when the length of the genes list is lower
#' than the value of the 'maxSplitSize' argument. So you should set it to 1
#' to have the maximum number of genes for the max change attack. This function
#' is used by the 'overMaxChange' function with a default argument value of 100
#' to increase speed, and still returns significant results.
#' @param exprs DelayedMatrix of numeric RNA expression, cells are rows and genes
#' are columns - or a SingleCellExperiment object, a matrix or a data.frame. By default,
#' these are converted to a data.frame to increase speed performance during modifications.
#' However, this conversion can consume a significant amount of memory, see 'argForModif'
#' argument for options.
#' @param clusters a character vector of the clusters to which the cells belong
#' @param target the name of the cluster to modify
#' @param classifier a classifier in the suitable format.
#' A classifier function should be formated as follow:
#' classifier = function(expr, clusters, target){
#'      # Making the classification
#'      c("cell type", score)
#' }
#' `score` should be numeric between 0 and 1, 1 being the highest confidance
#' into the cell type classification.
#' The matrix `expr` contains RNA expression values, the vector `clusters`
#' consists of the cluster IDs for each cell in `expr`, and `target` is the
#' ID of the cluster for which we want to have a classification.
#' The function returns a vector with the classification result, and a score.
#' @param exclGenes a list of genes to exclude from the analysis
#' @param genes a list of genes in case you want to limit the
#' attack on a subset of genes
#' @param advMethod the name of the method to use
#' @param advFixedValue the numeric value to use in case of
#' advMethod=`fixed`
#' @param advFct the function to use in case advMethod
#' belongs to the following list: `full_row_fct`, `target_row_fct`,
#' `target_matrix_fct`, `full_matrix_fct`
#' @param maxSplitSize max size of dichotomic slices.
#' @param argForClassif the type of the first argument to feed to the
#' classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
#' or 'DelayedMatrix'.
#' @param argForModif type of matrix during for the modification, 'data.frame'
#' by default. Can be 'DelayedMatrix', which needs less memory but is slower.
#' @param verbose logical, set to TRUE to activate verbose mode
#' @return a character vector of genes you can modify on a cluster without
#' modifying its classification
#' @examples
#' library(DelayedArray)
#' 
#' MyClassifier <- function(expr, clusters, target) {
#'    c("T cell", 0.9)
#' }
#' rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
#'      CD8B=c(2,2,3,3)))
#' genes <- c("CD4", "CD8A")
#' clusters_id <- c("B cell","B cell","T cell","T cell")
#' 
#' advMaxChange(rna_expression, clusters_id,
#' "T cell", MyClassifier, advMethod="perc99")
#' @export
advMaxChange <- function(exprs, clusters, target, classifier,
                        exclGenes = c(), genes = c(), advMethod = "perc99",
                        advFixedValue = 3, advFct = NULL,
                        maxSplitSize = 1, argForClassif = 'data.frame',
                        argForModif = 'data.frame', slot=NULL,
                        verbose = FALSE) {
    if (!is(exprs, 'matrix') && !is(exprs,'data.frame') &&
        !is(exprs,'SingleCellExperiment') && !is(exprs,'DelayedMatrix') &&
        !is(exprs,'Seurat')){
        stop("The argument exprs must be a DelayedMatrix, a SingleCellExperiment, a matrix, a data.frame or a Seurat object.")
    }
    if (!is.character(clusters)) {
        stop("The argument clusters must be a vector of character.")
    }
    if (!is.character(target)) {
        stop("The argument target must be character.")
    }
    if (!is.function(classifier)){
        stop("The argument classifier must be a function.")
    }
    if (!is.character(exclGenes) && length(exclGenes)>0) {
        stop("The argument exclGenes must be character or vector of character.")
    }
    if (!is.character(genes) && length(genes)>0) {
        stop("The argument genes must be character or vector of character.")
    }
    if (!is.character(advMethod)) {
        stop("The argument advMethod must be character.")
    }
    if (!is.numeric(maxSplitSize)){
        stop("The argument maxSplitSize must be numeric.")
    }
    if (!is.character(argForClassif)) {
        stop("The argument argForClassif must be character: 'data.frame' or 'SingleCellExperiment'.")
    }
    if (!is.logical(verbose)){
        stop("The argument verbose must be logical.")
    }
    if (is(exprs,'SingleCellExperiment') ){
        exprs <- t(as.matrix(counts(exprs)))
    }
    if (!is(exprs,'DelayedMatrix') && argForModif=="DelayedMatrix"){
        message("Converting exprs object to a DelayedArray object")
        exprs <- DelayedArray::DelayedArray(exprs)
    }
    if (!is(exprs,'data.frame') && argForModif=="data.frame"){
        message("Converting exprs object to a data.frame object")
        exprs <- as.data.frame(exprs)
    }

    if (maxSplitSize < 1){ maxSplitSize <- 1 }
    genesToKeep <- unlist(lapply(unique(genes), function(strGene){
        if (!is.na(match(strGene, colnames(exprs)))) {
            return(strGene)
        } else {
            return(NULL)
        }
    }))
    if (length(genes) == 0) {
        if (tolower(argForClassif) == "seurat"){
            genesToKeep <- rownames(exprs@assays$RNA@features)
        } else {
            genesToKeep <- colnames(exprs)
        }
    }
    for (strGene in unique(exclGenes)) {
        if (!is.na(match(strGene, genesToKeep))) {
            genesToKeep <- genesToKeep[-match(strGene, genesToKeep)]
        }
    }

    modObj <- predictWithNewValue(exprs, c("1"),
            clusters, target, classifier,
            advMethod = "none",
            argForClassif = argForClassif, argForModif = argForModif,
            slot=slot, verbose = verbose)
    classifTarget <- unlist(unname(modObj[1]))

    lResults <- .predictionDichotMaxChange(exprs, genesToKeep,
        clusters, target, classifier,
        advMethod = advMethod,
        advFixedValue = advFixedValue, advFct = advFct,
        maxSplitSize = maxSplitSize, argForClassif = argForClassif,
        argForModif=argForModif, slot=slot,classifTarget=classifTarget,
        verbose = verbose
    )
    if ( is.null(lResults)){
        NULL
    } else {
        new("advChar", values=lResults)
    }
}

.dichotMaxSameType <- function(lResults, cGeneSplitValue, exprs,
                clusters, target, classifier, advMethod,
                advFixedValue, advFct, maxSplitSize, argForClassif, argForModif,
                slot,classifTarget, verbose){
        if (verbose) { message("same cellType") }
        if (length(lResults) == 0) {
            lResults <- cGeneSplitValue
        } else {
            if (verbose){
                message("check if concat lists still gives target")
            }
            concatGenesList <- c(lResults, cGeneSplitValue)
            if (verbose) {
                message(paste0(
                    "length of tested gene list: ",
                    length(unique(concatGenesList))
                ))
            }
            # check if concat lists still gives target
            modObj <- predictWithNewValue(exprs, concatGenesList, clusters,
                target, classifier,
                advMethod = advMethod,
                advFixedValue = advFixedValue, advFct = advFct,
                argForClassif = argForClassif, argForModif=argForModif,
                slot=slot, verbose = verbose
            )
            concatCellType <- modObj[1]
            if (concatCellType == classifTarget) {
                if (verbose) { message("YES: merge results") }
                # Merge results
                lResults <- c(lResults, cGeneSplitValue)
                if (verbose) {
                    message(paste0( "result length after merge: ", length(lResults)))
                } else {
                    message(paste0( "result length: ", length(lResults)))
                }
            } else {
                if (verbose) { message("NO: split and retry") }
                if (length(cGeneSplitValue) > maxSplitSize) {
                    lResults <- .predictionDichotMaxChange(exprs,
                        cGeneSplitValue, clusters, target,
                        classifier,
                        lResults = lResults,
                        advMethod = advMethod,
                        advFixedValue = advFixedValue,
                        advFct = advFct,
                        maxSplitSize = maxSplitSize,
                        argForClassif = argForClassif,
                        argForModif=argForModif,
                        slot = slot,
                        classifTarget=classifTarget,
                        verbose = verbose
                    )
                }
            }
        }
    lResults
}

.dichotMaxSplit <- function(lResults, cGeneSplitValue, exprs,
                    genes, clusters, target,
                                    classifier,
                                    advMethod,
                                    advFixedValue,
                                    advFct, maxSplitSize,
                                    argForClassif,
                                    argForModif,
                                    slot,
                                    classifTarget,
                                    verbose){
    if (length(cGeneSplitValue) != 0) {
        if (verbose) {
            message("before predictWithNewValue 1")
            message(length(cGeneSplitValue))
        }
        modObj <- predictWithNewValue(exprs, cGeneSplitValue,
            clusters, target, classifier,
            advMethod = advMethod,
            advFixedValue = advFixedValue, advFct = advFct,
            argForClassif = argForClassif,
            argForModif=argForModif,
            slot=slot,
            verbose = verbose
        )
        cellType <- modObj[1]
        celltypeScore <- modObj[2]
        if (verbose){
            message(cellType)
            message(celltypeScore)
        }
    }
    if (cellType == classifTarget) {
        lResults <- .dichotMaxSameType(lResults,
                    cGeneSplitValue,
                    exprs, clusters,
                    target, classifier,
                    advMethod,
                    advFixedValue, advFct,
                    maxSplitSize,
                    argForClassif,
                    argForModif,
                    slot,
                    classifTarget,
                    verbose)
    } else {
        if (verbose) { message("NOT same cellType") }
        if (length(cGeneSplitValue) > maxSplitSize) {
            lResults <- .predictionDichotMaxChange(exprs,
                unname(cGeneSplitValue), clusters, target,
                classifier, lResults = lResults, advMethod = advMethod,
                advFixedValue = advFixedValue, advFct = advFct,
                maxSplitSize = maxSplitSize, argForClassif = argForClassif,
                argForModif=argForModif, slot=slot,
                classifTarget=classifTarget, verbose = verbose
            )
        }
    }
    lResults
}

.custSplit <- function(v1, v2){
    result <- list()
    for ( i in seq_along(v1)){
        strKey = as.character(v2[(i %% length(v2)) + 1])
        if ( !strKey %in% names(result)){
            result[[strKey]] = v1[i]
        } else {
            result[[strKey]] = c(result[[strKey]], v1[i])
        }
    }
    result
}

.predictionDichotMaxChange <- function(exprs, genes, clusters, target,
                                    classifier, lResults = c(),
                                    advMethod = "perc99",
                                    advFixedValue = 3,
                                    advFct = NULL,
                                    maxSplitSize = 1,
                                    argForClassif,
                                    argForModif,
                                    slot,
                                    classifTarget,
                                    verbose) {
    cGeneSplit <- .custSplit(unlist(genes), c(1,2))
    if (verbose){
        message(paste0("genes size: ", length(unlist(genes))))
        message(paste0("current gene results length: ", length(lResults)))
    }
    # Random order for each reccursion
    randomIndex <- sample(c(1,2), 2)
    cGeneSplit <- cGeneSplit[randomIndex]
    lResults <- .dichotMaxSplit(lResults, unname(unlist(cGeneSplit[1])),
                    exprs, genes, clusters, target, classifier, advMethod,
                    advFixedValue, advFct, maxSplitSize,
                    argForClassif, argForModif, 
                    slot, classifTarget, verbose)
    lResults <- .dichotMaxSplit(lResults, unname(unlist(cGeneSplit[2])),
                exprs, genes, clusters, target, classifier, advMethod,
                advFixedValue, advFct, maxSplitSize,
                argForClassif, argForModif,
                slot, classifTarget, verbose)
    lResults
}

#' Returns a classification and an odd value from a
#' RNA expression DelayedMatrix or a SingleCellExperiment object,
#' for given genes, for a given cluster, for a given modification.
#'
#' @details This function aims to concatenate the following actions:
#' - modify the RNA gene expression
#' - classify the result
#' This is a widely used function in the other functions of the package.
#' @param exprs DelayedMatrix of numeric RNA expression, cells are rows and genes
#' are columns - or a SingleCellExperiment object, a matrix or a data.frame.
#' @param genes the character vector of genes to modify
#' @param clusters a character vector of the clusters to which the cells belong
#' @param target the name of the cluster to modify
#' @param classifier a classifier in the suitable format.
#' A classifier function should be formated as follow:
#' classifier = function(expr, clusters, target){
#'      # Making the classification
#'      c("cell type", score)
#' }
#' `score` should be numeric between 0 and 1, 1 being the highest confidance
#' into the cell type classification.
#' The matrix `expr` contains RNA expression values, the vector `clusters`
#' consists of the cluster IDs for each cell in `expr`, and `target` is the
#' ID of the cluster for which we want to have a classification.
#' The function returns a vector with the classification result, and a score.
#' @param advMethod the name of the method to use
#' @param advFixedValue the numeric value to use in case of
#' advMethod=`fixed`
#' @param advFct the function to use in case advMethod
#' belongs to the following list: `full_row_fct`, `target_row_fct`,
#' `target_matrix_fct`, `full_matrix_fct`
#' @param argForClassif the type of the first argument to feed to the
#' classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
#' or 'DelayedMatrix'.
#' @param argForModif type of matrix during for the modification, 'data.frame'
#' by default. Can be 'DelayedMatrix', which is slower, but need less memory.
#' @param verbose logical, set to TRUE to activate verbose mode
#' @return a vector of the classification, and the associated odd
#' @examples
#' library(DelayedArray)
#' 
#' MyClassifier <- function(expr, clusters, target) {
#'    c("T cell", 0.9)
#' }
#' rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0),
#'      CD8A=c(1,1,1,1),
#'      CD8B=c(2,2,3,3)))
#' genes <- c("CD4", "CD8A")
#' clusters_id <- c("B cell","B cell","T cell","T cell")
#'
#' predictWithNewValue(rna_expression, genes, clusters_id,
#' "T cell", MyClassifier, advMethod="perc99")
#' @export
predictWithNewValue <- function(exprs, genes, clusters, target,
                                classifier, advMethod = "perc99",
                                advFixedValue = 3,
                                advFct = NULL,
                                argForClassif = 'data.frame',
                                argForModif = 'data.frame',
                                slot=NULL,
                                verbose = FALSE) {
    if (!is(exprs, 'matrix') && !is(exprs,'data.frame') &&
        !is(exprs,'SingleCellExperiment') && !is(exprs,'DelayedMatrix') &&
        !is(exprs,'Seurat')){
        stop("The argument exprs must be a DelayedMatrix, a SingleCellExperiment, a matrix or a data.frame")
    }
    if (!is.character(genes)) {
        stop("The argument genes must be character or vector of character.")
    }
    if (!is.character(clusters)) {
        stop("The argument clusters must be a vector of character.")
    }
    if (!is.character(target)) {
        stop("The argument target must be character.")
    }
    if (!is.function(classifier)){
        stop("The argument classifier must be a function.")
    }
    if (!is.character(advMethod)) {
        stop("The argument advMethod must be character.")
    }
    if (!is.character(argForClassif)) {
        stop("The argument argForClassif must be character: 'data.frame' or 'SingleCellExperiment'.")
    }
    if (!is.logical(verbose)){
        stop("The argument verbose must be logical.")
    }

    if (verbose) {
        message("Modify data for ", length(genes), " genes for cluster ", target)
    }

                message("predictWithNewValue ", argForModif, " ", argForClassif)
    modifExprs <- advModifications(exprs,
        genes,
        clusters,
        target,
        advMethod = advMethod,
        advFixedValue = advFixedValue,
        advFct = advFct,
        argForClassif = argForClassif,
        argForModif = argForModif,
        slot = slot
    )

    if(verbose){
        message(class(modifExprs))
    }

    classifier(modifExprs, clusters, target)
}

title: "adverSCarial, generate and analyze the vulnerability of scRNA-seq
    classifiers to adversarial attacks"
shorttitle: "adverSCarial"
author: Ghislain FIEVET <ghislain.fievet@gmail.com>
package: adverSCarial
abstract: >
    adverSCarial is an R Package designed for generating and analyzing the vulnerability of scRNA-seq
    classifiers to adversarial attacks. The package is versatile and provides a format for integrating
    any type of classifier. It offers functions for studying and generating two types of attacks,
    min change attack and max change attack. The single gene attack involves making a small modification
    to the input to alter the classification. The max change attack involves making a large modification
    to the input without changing its classification.
    The package provides a comprehensive solution for evaluating the robustness of scRNA-seq classifiers
    against adversarial attacks.
vignette: >
  %\VignetteIndexEntry{Vign02_overView_analysis}
  %\VignetteEngine{knitr::knitr}
  %\VignetteEncoding{UTF-8}

# Load data
  
```{r load libraries, warning = FALSE, message=FALSE}
library(adverSCarial)
library(LoomExperiment)
library(DelayedArray)
```

## Load loom file
```{r min change attack, message=FALSE, warning = FALSE}
pbmcPath <- system.file("extdata", "pbmc_short.loom", package="adverSCarial")
lfile <- import(pbmcPath, type="SingleCellLoomExperiment")
```

```{r load table}
matPbmc <- counts(lfile)
```

```{r visualize1, message=FALSE, warning = FALSE}
matPbmc[1:5,1:5]
```
## Load cell type annotations
```{r Load cell type annotations, message=FALSE, warning = FALSE}
cellTypes <- rowData(lfile)$cell_type
```

```{r visualize2, message=FALSE, warning = FALSE}
head(cellTypes)
```

# Run vulnerability analysis with `singleGeneOverview` and `maxChangeOverview`

The `singleGeneOverview` and `maxChangeOverview` functions are designed to provide insight into of the min and max change adversarial attacks on each cell type, on various gene modifications.

## Which attack to choose?
Before generating an attack it is judicious to choose the cell type to attack, and the modification susceptible to lead to a successful attack. Both functions run attack approximations, faster than the original, by studying splices of 100 genes.

## Which classifier is more vulnerable to adversarial attacks?
Sometimes we want to compare two classifiers and see which one is more vulnerable to adversarial attacks.


## Which modifications to compare

Here we define the modifications to analyse, the predefined `perc1`, and a custom function returning high outliers called `modifOutlier`.
```{r modifications for overview, warning = FALSE}
modifOutlier <- function(x, y){
    return (max(x)*1000)
}
```

```{r use the function, warning = FALSE}
modifications <- list(c("perc1"), c("full_row_fct", modifOutlier))
```

We run the `singleGeneOverview`, this gives us a general idea of which cell types are more vulnerable to single gene attacks.
```{r run min change overview, warning = FALSE, message=FALSE}
min_change_overview <- singleGeneOverview(matPbmc, cellTypes, MClassifier,
    modifications= modifications, maxSplitSize = 20, firstDichot = 5)
```
```{r run min change overview bis, warning = FALSE}
min_change_overview
```

And the `maxChangeOverview`, giving us a general idea of which cell types are more vulnerable to max change attacks.
```{r run max change overview, warning = FALSE, message=FALSE}
max_change_overview <- maxChangeOverview(matPbmc, cellTypes, MClassifier,
    modifications= modifications, maxSplitSize = 20)
```
```{r run max change overview bis, warning = FALSE}
max_change_overview
```

```{r session info}
sessionInfo()
```

---
title: "adverSCarial, generate and analyze the vulnerability of scRNA-seq
    classifiers to adversarial attacks"
shorttitle: adverSCarial
author: Ghislain FIEVET <ghislain.fievet@gmail.com>
package: adverSCarial
abstract: >
    adverSCarial is an R Package designed for generating and analyzing the vulnerability of scRNA-seq
    classifiers to adversarial attacks. The package is versatile and provides a format for integrating
    any type of classifier. It offers functions for studying and generating two types of attacks,
    single gene attack and max change attack. The single gene attack involves making a small modification
    to the input to alter the classification. The max change attack involves making a large modification
    to the input without changing its classification.
    The package provides a comprehensive solution for evaluating the robustness of scRNA-seq classifiers
    against adversarial attacks.
output:
    BiocStyle::html_document:
        toc: true
        toc_depth: 2
vignette: >
  %\VignetteIndexEntry{Vign01_adverSCarial}
  %\VignetteEngine{knitr::knitr}
  %\VignetteEncoding{UTF-8}
---

# Introduction
**adverSCarial is a package for generating and evaluating vulnerability to adversarial attacks on single-cell RNA sequencing classifiers.**
Single cell analysis are on the way to be used in clinical routine. As a critical use, algorithms must address the challenge of ensuring reliability, one concern being the susceptibility to adversarial attacks.


Zhang, J., Wang, W., Huang, J., Wang, X., & Zeng, Y. (2020). How far is single-cell sequencing from clinical application?. Clinical and translational medicine, 10(3), e117. https://doi.org/10.1002/ctm2.117

de Hond, A.A.H., Leeuwenberg, A.M., Hooft, L. et al. Guidelines and quality criteria for artificial intelligence-based prediction models in healthcare: a scoping review. npj Digit. Med. 5, 2 (2022). https://doi.org/10.1038/s41746-021-00549-7


The package is designed to generate and analyze the vulnerability of scRNA-seq classifiers to adversarial attacks. The package is versatile and provides a format for integrating
any type of classifier. It offers functions for studying and generating two types of attacks, single gene attack and max change attack.
The single gene attack involves making a small modification to the input to alter the classification. This is an issue when the change of classification is made on a gene that is not known to be biologicaly involved in the change of this cell type.
The max change attack involves making a large modification to the input without changing its classification. This is an issue when a significant percentage of genes, sometimes as high as 99%, can be modified with the classifier still predicting the same cell type.

# Jupyter Notebook examples
- [Train, format and CGD attack on a scAnnotatR classifier](https://github.com/GhislainFievet/adverSCarial_article/blob/master/04_train_and_attack_scAnnotatR.ipynb)
- [Train, format and max-change attack on a random forest classifier](https://github.com/GhislainFievet/adverSCarial_article/blob/master/05_train_format_random_forest_classifier_scRF.ipynb)
- [Train, format and single-gene attack on a multi layer perceptron classifier](https://github.com/GhislainFievet/adverSCarial_article/blob/master/06_train_format_classifier_scMLP.ipynb)


# Installation
```{r installation, eval=FALSE}
## Install BiocManager is necessary
if (!require("BiocManager")) {
    install.packages("BiocManager")
}
BiocManager::install('adverSCarial')
```

# Generate an adversarial attack

There are three types of adversarial attacks: `single-gene attack` which modifies the expression of a gene in order to alter the classification, `max-change attack` which introduce the largest possible perturbations to the input while still keeping the same classification, and `Cluster Gradient Descent (CGD)` which modifies the input matrix gene by gene by following an approximated gradient until the cluster has switch its classification.

Load libraries

```{r load libraries, warning = FALSE, message=FALSE}
library(adverSCarial)
library(LoomExperiment)
```

Load a pbmc raw data

```{r min change attack, message=FALSE, warning = FALSE}
pbmcPath <- system.file("extdata", "pbmc_short.loom", package="adverSCarial")
lfile <- import(pbmcPath, type="SingleCellLoomExperiment")
```

```{r extract matrix data}
matPbmc <- counts(lfile)
```

```{r visualize1, message=FALSE, warning = FALSE}
matPbmc[1:5, 1:5]
```


and its cell type classification based on [Seurat documentation](https://satijalab.org/seurat/articles/pbmc3k_tutorial.html)
```{r load annnots, message=FALSE, warning = FALSE}
cellTypes <- rowData(lfile)$cell_type
```

```{r visualize2, message=FALSE, warning = FALSE}
head(cellTypes)
```

The package contains a marker based classifier working on the pbmc3k dataset: `MClassifier`.
We verify it is classifying properly according to the `cellTypes`.

```{r first classif, message=FALSE}
for ( cell_type in unique(cellTypes)){
    resClassif <- MClassifier(matPbmc, cellTypes, cell_type)
    print(paste0("Expected: ", cell_type, ", predicted: ", resClassif[1]))
}
```

We want to run attacks on the "DC" cluster but without modifying the genes used by human to make manual classification. We can find this list on [Seurat documentation](https://satijalab.org/seurat/articles/pbmc3k_tutorial.html)

```{r markers}
# Known markers for each cell type
markers <- c("IL7R", "CCR7", "CD14", "LYZ", "S100A4", "MS4A1", "CD8A", "FCGR3A", "MS4A7",
              "GNLY", "NKG7", "FCER1A", "CST3", "PPBP")
```

## Single-gene attack

Modify the value of a gene inside a cell cluster in order to change its classification.

The function `advSingleGene` runs a dichotomic search of one gene attacks, given the cluster to attack and a type of modification. We suggest two main modifications: `perc1` replacing the value of the gene by its first percentile, and `perc99` replacing the value of the gene by its 99th percentile. The `adv_fct` argument allows users to choose custom modifications of the gene.

The `excl_genes` argument allows users to exclude certain genes from being modified. Here we exclude the genes usually used as markers of specific cell types, as defined previously.

Computation times can be lengthy, we use the `return_first_found` argument to return the result as soon as it is found.

We can specify the `change_type = "not_na"` argument to indicate that we want the attack to misclassify the cluster as an existing cell type, rather than as NA.

```{r search min change attack, message=FALSE, warning = FALSE}
genesMinChange <- advSingleGene(matPbmc, cellTypes, "DC",
                        MClassifier, exclGenes = markers, advMethod = "perc99",
                        returnFirstFound = TRUE, changeType = "not_na",
                        firstDichot = 10)
```
```{r search min change attack bis, warning = FALSE}
genesMinChange
```

The function found that modifying the single gene `r names(genesMinChange@values)[1]` with the `perc99` modification on the cluster leads to a new classification, `r genesMinChange@values[[1]][1]`.

Let's run this attack and verify if it is successful.

First we modify the `matPbmc` matrix on the target cluster.

```{r run min change attack, message=FALSE, warning = FALSE}
matAdver <- advModifications(matPbmc, names(genesMinChange@values)[1], cellTypes, "DC")
```

Then classify the "DC" cluster with `MClassifier`.
```{r verify the min change attack, warning = FALSE, message = FALSE}
resClassif <- MClassifier(matAdver, cellTypes, "DC")
```
```{r verify the min change attack bis, warning = FALSE}
resClassif
```

## Max-change attack

Modify the maximum number of genes inside a cell cluster without altering its classification.

The function `advMaxChange` runs a dichotomic search of gene subsets, given the cluster to attack and a type of modification, such that the classification does not change.

The `maxSplitSize` argument is the maximum size of dichotomic slices. Set to 1 to have better results, but it will take longer to compute.

```{r search max change attack, warning = FALSE, message=FALSE}
genesMaxChange <- advMaxChange(matPbmc, cellTypes, "Memory CD4 T", MClassifier,
                    exclGenes = markers, advMethod = "perc99")
```
```{r search max change attack bis, warning = FALSE}
length(genesMaxChange@values)
```


The function found `r length(genesMaxChange@values)` genes that you can modify with the `perc99` modification, and the cluster is still classified as `Memory CD4 T`.

Let's run this attack and verify if it is successful.

First we modify the `matPbmc` matrix on the target cluster, on the genes previously determined.

```{r run max change attack, message=FALSE, warning=FALSE}
matMaxAdver <- advModifications(matPbmc, genesMaxChange@values, cellTypes, "Memory CD4 T")
```

Then we verify that classification is still `Memory CD4 T`.
```{r verify max change attack, warning = FALSE, message=FALSE}
resClassif <- MClassifier(matMaxAdver, cellTypes, "Memory CD4 T")
```
```{r verify max change attack bis, warning = FALSE}
resClassif
```


## CGD attack

Apply the Cluster Gradient Descent (CGD) attack on the "Memory CD4 T" cells cluster with parameters alpha=epsilon=1

```{r CGD1, warning = FALSE}
resCGD <- advCGD(as.data.frame(matPbmc), cellTypes,
                "Memory CD4 T", MClassifier, alpha=1, epsilon=1,
                genes=colnames(matPbmc)[ncol(matPbmc):1])
```

The algorithm test each gene and stops when the cluster prediction has switched
to another cell type prediction. Here it stops after having modified the CCR5 gene which make the classifier switch its prediction to Naive CD4 T.

```{r CGD2, warning = FALSE}
tail(resCGD$byGeneSummary)
```
```{r CGD2_bis, warning = FALSE}
resCGD$oneRowSummary
```
You can retrieve the modified matrix.

```{r CGD3, warning = FALSE}
modifiedMatrix <- resCGD$expr
```

And visualize the list of modified genes.

```{r CGD4, warning = FALSE}
resCGD$modGenes
```

Check the new classification of the modified matrix.

```{r CGD5, warning = FALSE}
MClassifier(modifiedMatrix, cellTypes, "Memory CD4 T")$prediction
```

```{r session info}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advModifications.R
\name{advModifications}
\alias{advModifications}
\title{Returns a modified RNA expression DelayedMatrix, or a modified SingleCellExperiment,
for a given cluster, for a given modification.}
\usage{
advModifications(
  exprs,
  genes,
  clusters,
  target,
  advMethod = "perc99",
  advFixedValue = 3,
  advFct = NULL,
  argForClassif = "DelayedMatrix",
  argForModif = "data.frame",
  slot = NULL,
  verbose = FALSE
)
}
\arguments{
\item{exprs}{DelayedMatrix of numeric RNA expression, cells are rows and genes
are columns - or a SingleCellExperiment object, a matrix or a data.frame. By default,
these are converted to a data.frame to increase speed performance during modifications.
However, this conversion can consume a significant amount of memory, see 'argForModif'
argument for options.}

\item{genes}{the character vector of genes to modify}

\item{clusters}{a character vector of the clusters to which the cells belong}

\item{target}{the name of the cluster to modify}

\item{advMethod}{the name of the method to use}

\item{advFixedValue}{the numeric value to use in case of
advMethod=\code{fixed}}

\item{advFct}{the function to use in case advMethod
belongs to the following list: \code{full_row_fct}, \code{target_row_fct},
\code{target_matrix_fct}, \code{full_matrix_fct}}

\item{argForClassif}{the type of the first argument to feed to the
classifier function. 'DelayedMatrix' by default, can be 'SingleCellExperiment'
or 'data.frame'.}

\item{argForModif}{type of matrix during for the modification, 'DelayedMatrix'
by default. Can be 'data.frame', which is faster, but need more memory.}

\item{verbose}{logical, set to TRUE to activate verbose mode}
}
\value{
the matrix or a data.frame exprs modified on asked genes
with the specified modification
}
\description{
Returns a modified RNA expression DelayedMatrix, or a modified SingleCellExperiment,
for a given cluster, for a given modification.
}
\details{
The motivation for this function is to standardize the modifications
we want to study in the attacks. We give as argument a DelayedMatrix of the RNA
expression, the gene and the target cells we want to modify. Then we have
three arguments allowing to specify what modification we want to apply on
these cells. The advMethod contains, a specific prerecorded modification
or an indication on how to use the other two arguments.
The prerecorded modifications available for the advMethod argument are:
\itemize{
\item 'perc1', replace the value by the whole matrix 1 percentile value of
the gene. It is as if we biologically switched off the gene.
\item 'perc99', replace the value by the whole matrix 99 percentile value of
the gene. It is as if we biologically switched on the gene to the maximum.
\item 'random', replace the value by from a uniform distribution between min
and max of the gene on the dataset
\item 'positive_aberrant' replace value by 10,000 times the max value of the
gene on the dataset
\item 'negative_aberrant' replace value by -10,000 times the max value of the
gene on the dataset
\item 'decile+X', shifts the gene value by + X deciles.
\item 'decile-X', shifts the gene value by - X deciles.
The value of the advMethod argument can also be 'fixed', in this case the
modification would be to replace the value of the gene of the wanted cells
by the value of the argument 'advFixedValue'. This can be useful to test
aberrant values like negative integer, absurdly high values of character
values.
The value of the advMethod argument can also be 'full_row_fct',
'target_row_fct', 'target_matrix_fct' or 'full_matrix_fct'. They are
used when we want to use a custom modification function, with the 'advFct'
argument:
\item 'full_row_fct' indicate that the 'advFct' function takes the whole gene
values as input.
\item 'target_row_fct' indicate that the 'advFct' function takes target cells
gene values as input.
\item 'full_matrix_fct' indicate that the 'advFct' function takes the whole gene
expression values as input.
\item 'target_matrix_fct' indicate that the 'advFct' function takes target cells
all genes values as input.
}
}
\examples{
library(DelayedArray)

rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
     CD8B=c(2,2,3,3)))
genes <- c("CD4", "CD8A")
clusters_id <- c("B cell","B cell","T cell","T cell")

advModifications(rna_expression, genes, clusters_id,
"T cell", advMethod="perc99")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advGridMinChange.R
\name{advGridMinChange}
\alias{advGridMinChange}
\title{Grid search of min change adversarial attack. Tries each
combination on a cluster, given a list of genes and a list of modifications.}
\usage{
advGridMinChange(
  exprs,
  clusters,
  target,
  classifier,
  genes,
  modifications = list(c("perc1"), c("perc99")),
  returnFirstFound = FALSE,
  argForClassif = "data.frame",
  argForModif = "data.frame",
  verbose = FALSE,
  iamsure = FALSE
)
}
\arguments{
\item{exprs}{DelayedMatrix of numeric RNA expression, cells are rows and genes
are columns - or a SingleCellExperiment object, a matrix or a data.frame. By
default matrix and data.frame are converted to DelayedMatrix for memory
performance, see 'argForModif' argument for options.}

\item{clusters}{a character vector of the clusters to which the cells belong}

\item{target}{the name of the cluster to modify}

\item{classifier}{a classifier in the suitable format.
A classifier function should be formated as follow:
classifier = function(expr, clusters, target){
# Making the classification
c("cell type", score)
}
\code{score} should be numeric between 0 and 1, 1 being the highest confidance
into the cell type classification.
The matrix \code{expr} contains RNA expression values, the vector \code{clusters}
consists of the cluster IDs for each cell in \code{expr}, and \code{target} is the
ID of the cluster for which we want to have a classification.
The function returns a vector with the classification result, and a score.}

\item{genes}{the character vector of genes to study}

\item{modifications}{the list of the modifications to study}

\item{returnFirstFound}{set to TRUE to return result when a
the first misclassification is found}

\item{argForClassif}{the type of the first argument to feed to the
classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
or 'DelayedMatrix'.}

\item{argForModif}{type of matrix during for the modification, 'data.frame'
by default. Can be 'DelayedMatrix', which needs less memory but is slower.}

\item{verbose}{logical, set to TRUE to activate verbose mode}

\item{iamsure}{logical, prevents from expansive calculations
when \code{genes} list is too long, set to \code{TRUE} to run anyway.}
}
\value{
DataFrame results of the classification of all the grid combinations
}
\description{
Grid search of min change adversarial attack. Tries each
combination on a cluster, given a list of genes and a list of modifications.
}
\details{
This function aims to find the shortest combination of genes
allowing to make a min change attack. It will test every possible combination
for a given gene list. This function can take a long time to run, and we
recommand to use the random walk search advRandWalkMinChange function instead
for lists above 10 genes.

You can specify a list of modifications as so, each item of the list should be 1 or 2 length size.
The 1 length vector must contain the prerecorded modifications, 'perc1' or 'perc99'.
The 2 length vector must have as first item:
\itemize{
\item 'fixed', in this case the second item should be the value to be
replaced by.
\item 'full_row_fct', 'target_row_fct', 'target_matrix_fct' or
'full_matrix_fct'. In this case the second item should be a function.
Let's say we want to analysis the susceptibility to min change attack
for 3 modifications: "perc1", the modification of each value of the
cluster by 1000, and a custom modification stored inside a function myFct.
Then the 'modification' parameter should be:
my_modifications = list(c("perc1"),
c("fixed", 1000),
c("full_matrix_fct", myFct))
}
}
\examples{
library(DelayedArray)

MyClassifier <- function(expr, clusters, target) {
   c("T cell", 0.9)
}
rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
     CD8B=c(2,2,3,3)))
genes <- c("CD4", "CD8A")
clusters_id <- c("B cell","B cell","T cell","T cell")

advGridMinChange(rna_expression, clusters_id, "T cell",
 MyClassifier, genes=genes,
 modifications = list(c("perc1"), c("perc99")))

myModif = function(x, y){
   return(sample(1:10,1))
}

my_modifications = list(c("perc1"),
                        c("fixed", 1000),
                        c("full_matrix_fct", myModif))
advGridMinChange(rna_expression, clusters_id, "T cell",
 MyClassifier, genes=genes, modifications = my_modifications)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sceConvertToHGNC.R
\name{sceConvertToHGNC}
\alias{sceConvertToHGNC}
\title{Returns a SingleCellExperiment object keeping unique HGNC gene}
\usage{
sceConvertToHGNC(sce)
}
\arguments{
\item{sce}{SingleCellExperiment object to convert}
}
\value{
the SingleCellExperiment object keeping unique HGNC gene
}
\description{
Returns a SingleCellExperiment object keeping unique HGNC gene
}
\details{
Sometimes classifiers need HGNC instead of ensemble genes
to run. This function allows to make the conversion.
}
\examples{
library(TENxPBMCData)

pbmc <- TENxPBMCData(dataset = "pbmc3k")
hgnc_pbmc <- sceConvertToHGNC(pbmc)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictWithNewValue.R
\name{predictWithNewValue}
\alias{predictWithNewValue}
\title{Returns a classification and an odd value from a
RNA expression DelayedMatrix or a SingleCellExperiment object,
for given genes, for a given cluster, for a given modification.}
\usage{
predictWithNewValue(
  exprs,
  genes,
  clusters,
  target,
  classifier,
  advMethod = "perc99",
  advFixedValue = 3,
  advFct = NULL,
  argForClassif = "data.frame",
  argForModif = "data.frame",
  slot = NULL,
  verbose = FALSE
)
}
\arguments{
\item{exprs}{DelayedMatrix of numeric RNA expression, cells are rows and genes
are columns - or a SingleCellExperiment object, a matrix or a data.frame.}

\item{genes}{the character vector of genes to modify}

\item{clusters}{a character vector of the clusters to which the cells belong}

\item{target}{the name of the cluster to modify}

\item{classifier}{a classifier in the suitable format.
A classifier function should be formated as follow:
classifier = function(expr, clusters, target){
# Making the classification
c("cell type", score)
}
\code{score} should be numeric between 0 and 1, 1 being the highest confidance
into the cell type classification.
The matrix \code{expr} contains RNA expression values, the vector \code{clusters}
consists of the cluster IDs for each cell in \code{expr}, and \code{target} is the
ID of the cluster for which we want to have a classification.
The function returns a vector with the classification result, and a score.}

\item{advMethod}{the name of the method to use}

\item{advFixedValue}{the numeric value to use in case of
advMethod=\code{fixed}}

\item{advFct}{the function to use in case advMethod
belongs to the following list: \code{full_row_fct}, \code{target_row_fct},
\code{target_matrix_fct}, \code{full_matrix_fct}}

\item{argForClassif}{the type of the first argument to feed to the
classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
or 'DelayedMatrix'.}

\item{argForModif}{type of matrix during for the modification, 'data.frame'
by default. Can be 'DelayedMatrix', which is slower, but need less memory.}

\item{verbose}{logical, set to TRUE to activate verbose mode}
}
\value{
a vector of the classification, and the associated odd
}
\description{
Returns a classification and an odd value from a
RNA expression DelayedMatrix or a SingleCellExperiment object,
for given genes, for a given cluster, for a given modification.
}
\details{
This function aims to concatenate the following actions:
\itemize{
\item modify the RNA gene expression
\item classify the result
This is a widely used function in the other functions of the package.
}
}
\examples{
library(DelayedArray)

MyClassifier <- function(expr, clusters, target) {
   c("T cell", 0.9)
}
rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0),
     CD8A=c(1,1,1,1),
     CD8B=c(2,2,3,3)))
genes <- c("CD4", "CD8A")
clusters_id <- c("B cell","B cell","T cell","T cell")

predictWithNewValue(rna_expression, genes, clusters_id,
"T cell", MyClassifier, advMethod="perc99")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getSignGenes.R
\name{getSignGenes}
\alias{getSignGenes}
\title{The function getSignGenes orders the genes by maximizing the
significance of the gene to differentiate the clusters and ensures
that they represent at most the variations across all possible pairs
of clusters.}
\usage{
getSignGenes(expr, clusters, method = "wilcox", verbose = FALSE)
}
\arguments{
\item{expr}{A matrix of gene expression data. Rows are cells and columns
are genes.}

\item{clusters}{a character vector of the clusters to which the cells belong}

\item{method}{the statistical test to use. Either "wilcox" for the Wilcoxon
rank sum test or "ttest" for the t-test. Default is "wilcox".}

\item{verbose}{logical, set to TRUE to activate verbose mode}
}
\description{
The function getSignGenes orders the genes by maximizing the
significance of the gene to differentiate the clusters and ensures
that they represent at most the variations across all possible pairs
of clusters.
}
\details{
The function getSignGenes orders the genes by maximizing the
significance of the gene to differentiate the clusters and ensures
that they represent at most the variations across all possible pairs
of clusters.
}
\examples{
rna_expression <- data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
    CD8B=c(2,2,3,3))
genes <- c("CD4", "CD8A")
clusters_id <- c("B cell","T cell","T cell","T cell")

getSignGenes(rna_expression, clusters_id, method="wilcox", verbose=TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advMaxChange.R
\name{advMaxChange}
\alias{advMaxChange}
\title{Find a max change adversarial attack. It finds the longer
list of genes you can modify on a cluster without changing its
classification.}
\usage{
advMaxChange(
  exprs,
  clusters,
  target,
  classifier,
  exclGenes = c(),
  genes = c(),
  advMethod = "perc99",
  advFixedValue = 3,
  advFct = NULL,
  maxSplitSize = 1,
  argForClassif = "data.frame",
  argForModif = "data.frame",
  slot = NULL,
  verbose = FALSE
)
}
\arguments{
\item{exprs}{DelayedMatrix of numeric RNA expression, cells are rows and genes
are columns - or a SingleCellExperiment object, a matrix or a data.frame. By default,
these are converted to a data.frame to increase speed performance during modifications.
However, this conversion can consume a significant amount of memory, see 'argForModif'
argument for options.}

\item{clusters}{a character vector of the clusters to which the cells belong}

\item{target}{the name of the cluster to modify}

\item{classifier}{a classifier in the suitable format.
A classifier function should be formated as follow:
classifier = function(expr, clusters, target){
# Making the classification
c("cell type", score)
}
\code{score} should be numeric between 0 and 1, 1 being the highest confidance
into the cell type classification.
The matrix \code{expr} contains RNA expression values, the vector \code{clusters}
consists of the cluster IDs for each cell in \code{expr}, and \code{target} is the
ID of the cluster for which we want to have a classification.
The function returns a vector with the classification result, and a score.}

\item{exclGenes}{a list of genes to exclude from the analysis}

\item{genes}{a list of genes in case you want to limit the
attack on a subset of genes}

\item{advMethod}{the name of the method to use}

\item{advFixedValue}{the numeric value to use in case of
advMethod=\code{fixed}}

\item{advFct}{the function to use in case advMethod
belongs to the following list: \code{full_row_fct}, \code{target_row_fct},
\code{target_matrix_fct}, \code{full_matrix_fct}}

\item{maxSplitSize}{max size of dichotomic slices.}

\item{argForClassif}{the type of the first argument to feed to the
classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
or 'DelayedMatrix'.}

\item{argForModif}{type of matrix during for the modification, 'data.frame'
by default. Can be 'DelayedMatrix', which needs less memory but is slower.}

\item{verbose}{logical, set to TRUE to activate verbose mode}
}
\value{
a character vector of genes you can modify on a cluster without
modifying its classification
}
\description{
Find a max change adversarial attack. It finds the longer
list of genes you can modify on a cluster without changing its
classification.
}
\details{
This function aims to get the largest part of the genes
that can be modified without altering the classification,
considering a given modification. You can refer to the
'advModifications' function documentation to more details on how
to define a modification.
The search is made by a dichotomic process, on a reccursive function.
At each iteration the function splits the genes in two groups. It
proceeds to the modification of the RNA gene value of the first group,
makes its classification. Then three possible scenarios:
\itemize{
\item the classification is the same as the target cluster. We concat
the genes list to the previous one, make the classification, and it still
gives same classification. Then we return the genes list.
\item the classification is the same as the target cluster. We concat
the genes list to the previous one, make the classification, and it
gives a different classification. This happens often, you can modify
the gene A with a classification of T cell, or modify the gene B with a
classification of T cell, but modifying A and B returns another
classification. In this case we split the genes list in two and try again.
\item the classification is not the same as the target cluster. In this case we
split the genes list in two and try again.
The iteration process stops when the length of the genes list is lower
than the value of the 'maxSplitSize' argument. So you should set it to 1
to have the maximum number of genes for the max change attack. This function
is used by the 'overMaxChange' function with a default argument value of 100
to increase speed, and still returns significant results.
}
}
\examples{
library(DelayedArray)

MyClassifier <- function(expr, clusters, target) {
   c("T cell", 0.9)
}
rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
     CD8B=c(2,2,3,3)))
genes <- c("CD4", "CD8A")
clusters_id <- c("B cell","B cell","T cell","T cell")

advMaxChange(rna_expression, clusters_id,
"T cell", MyClassifier, advMethod="perc99")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allClasses.R
\docType{class}
\name{advChar-class}
\alias{advChar-class}
\alias{advChar}
\title{adverSCarial class}
\value{
A advChar object
}
\description{
\code{advChar} is a class used to store the output values
of the \code{advMaxChange} function. The result can be a vector
of few thousands genes, so a specific \emph{show} method is associated to
this class to avoid overflooding the R scripts outputs.
}
\examples{
MyClassifier <- function(expr, clusters, target) {
  c("T cell", 0.9)
}

genes <- paste0("gene_",1:10000)
rna_expression <- data.frame(lapply(genes, function(x) numeric(0)))
rna_expression <- rbind(rna_expression, rep(1,10000))
rna_expression <- rbind(rna_expression, rep(2,10000))
colnames(rna_expression) <- genes
clusters_id <- c("B cell","T cell")

max_change_genes <- advMaxChange(rna_expression, clusters_id,
"T cell", MyClassifier, advMethod="perc99")
                                   
max_change_genes
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advRandWalkMinChange.R
\name{advRandWalkMinChange}
\alias{advRandWalkMinChange}
\title{Random walk search of min change adversarial attack.}
\usage{
advRandWalkMinChange(
  exprs,
  clusters,
  target,
  classifier,
  genes,
  modifications = list(c("perc1"), c("perc99")),
  firstBatch = 100,
  walkLength = 100,
  stepChangeRatio = 0.2,
  whileMaxCount = 10000,
  changeType = "any",
  argForClassif = "data.frame",
  argForModif = "data.frame",
  verbose = FALSE
)
}
\arguments{
\item{exprs}{DelayedMatrix of numeric RNA expression, cells are rows and genes
are columns - or a SingleCellExperiment object, a matrix or a data.frame. By
default matrix and data.frame are converted to DelayedMatrix for memory
performance, see 'argForModif' argument for options.}

\item{clusters}{a character vector of the clusters to which the cells belong}

\item{target}{the name of the cluster to modify}

\item{classifier}{a classifier in the suitable format.
A classifier function should be formated as follow:
classifier = function(expr, clusters, target){
# Making the classification
c("cell type", score)
}
\code{score} should be numeric between 0 and 1, 1 being the highest confidance
into the cell type classification.
The matrix \code{expr} contains RNA expression values, the vector \code{clusters}
consists of the cluster IDs for each cell in \code{expr}, and \code{target} is the
ID of the cluster for which we want to have a classification.
The function returns a vector with the classification result, and a score.}

\item{genes}{the character vector of genes to study}

\item{modifications}{the list of the modifications to study}

\item{firstBatch}{the maximum number of try in step 1}

\item{walkLength}{the maximum number of try in step 2}

\item{stepChangeRatio}{ratio of parameters change in new walk step}

\item{whileMaxCount}{the maximum number of try when looking
for new combination of parameters}

\item{changeType}{\code{any} consider each misclassification,
\code{not_na} consider each misclassification but NA.}

\item{argForClassif}{the type of the first argument to feed to the
classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
or 'DelayedMatrix'.}

\item{argForModif}{type of matrix during for the modification, 'data.frame'
by default. Can be 'DelayedMatrix', which needs less memory but is slower.}

\item{verbose}{logical, set to TRUE to activate verbose mode}
}
\value{
DataFrame results of the classification of all the grid combinations
}
\description{
Random walk search of min change adversarial attack.
}
\details{
The parameter search by grid can take a long time,
this function aims to make a parameter search faster.
We have a function, advSingleGene, looking for one gene attacks.
The advRandWalkMinChange function aims to find a multiple genes
attack, with the fewer genes possible.
At first the user have to provide a list of genes to test, for
example by running differential statistics between two cell clusters.
The user should also provide a list of modifications to test, to
define as so - each item of the list should be 1 or 2 length size.
The 1 length vector must contain the prerecorded modifications, 'perc1'
or 'perc99'.
The 2 length vector must have as first item:
\itemize{
\item 'fixed', in this case the second item should be the value to be
replaced by.
\item 'full_row_fct', 'target_row_fct', 'target_matrix_fct' or
'full_matrix_fct'. In this case the second item should be a function.
Let's say we want to analysis the susceptibility to min change attack
for 3 modifications: "perc1", the modification of each value of the
cluster by 1000, and a custom modification stored inside a function myFct.
Then the 'modification' parameter should be:
my_modifications = list(c("perc1"),
c("fixed", 1000),
c("full_matrix_fct", myFct))
}

Then the function will try to find the best combination of these genes
and modifications to make the min change attack.
Step 1 is to find a seed by trying random combinations of
genes and modifications on a cluster until the classification is altered.
Step 2 is to perform a random walk search to reduce the number of genes
needed to change the classification.
The
}
\examples{
library(DelayedArray)

MyClassifier <- function(expr, clusters, target) {
   c("T cell", 0.9)
}
rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
     CD8B=c(2,2,3,3)))
genes <- c("CD4", "CD8A")
clusters_id <- c("B cell","B cell","T cell","T cell")

advRandWalkMinChange(rna_expression, clusters_id, "T cell",
MyClassifier, genes=genes,
modifications = list(c("perc1"), c("perc99")))

# Stop at first attack discovery, whitout going into the walk
# parameter search.
advRandWalkMinChange(rna_expression, clusters_id, "T cell",
 MyClassifier, genes=genes,
 modifications = list(c("perc1"), c("perc99")), walkLength=0)

myModif = function(x, y){
   return(sample(1:10,1))
}

my_modifications = list(c("perc1"),
                        c("fixed", 1000),
                        c("full_matrix_fct", myModif))
advRandWalkMinChange(rna_expression, clusters_id, "T cell",
 MyClassifier, genes=genes, modifications = my_modifications)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matrixFromSCE.R
\name{matrixFromSCE}
\alias{matrixFromSCE}
\title{Returns the RNA expression matrix from a SingleCellExperiment
with unique hgnc gene names in columns}
\usage{
matrixFromSCE(sce)
}
\arguments{
\item{sce}{SingleCellExperiment object to convert}
}
\value{
the RNA expression matrix from a SingleCellExperiment
with unique hgnc gene names in columns
}
\description{
Returns the RNA expression matrix from a SingleCellExperiment
with unique hgnc gene names in columns
}
\details{
This function retrieves from a SingleCellExperiment
object the raw RNA expression value corresponding to the hgnc
genes. The resulting matrix can then be used with adverSCarial
packages.
}
\examples{
library(TENxPBMCData)

pbmc <- TENxPBMCData(dataset = "pbmc3k")
mat_rna <- matrixFromSCE(pbmc)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advCGD.R
\name{advCGD}
\alias{advCGD}
\title{Implementation of the IDG4C algorithm.}
\usage{
advCGD(
  expr,
  clusters,
  target,
  classifier,
  genes = NULL,
  exclNewTargets = NULL,
  newTarget = NULL,
  alpha = 0.1,
  epsilon = 0,
  slot = NULL,
  stopAtSwitch = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{expr}{a matrix, a data.fram or a Seurat object}

\item{clusters}{a character vector of the clusters to which the cells belong}

\item{target}{the name of the cluster to modify}

\item{classifier}{a classifier in the suitable format.
A classifier function should be formated as follow:
classifier = function(expr, clusters, target){
# Making the classification
c("cell type", score)
}
\code{score} should be numeric between 0 and 1, 1 being the highest confidance
into the cell type classification.
The matrix \code{expr} contains RNA expression values, the vector \code{clusters}
consists of the cluster IDs for each cell in \code{expr}, and \code{target} is the
ID of the cluster for which we want to have a classification.
The function returns a vector with the classification result, and a score.}

\item{genes}{the character vector of genes to study}

\item{exclNewTargets}{the character vector of cell types to exclude as new target}

\item{newTarget}{the name of the new cell type cell type target, this will be the
new prediction cell type after the attack}

\item{alpha}{the alpha parameter of the IDG4C algorithm}

\item{epsilon}{the epsilon parameter of the IDG4C algorithm}

\item{slot}{the slot to modify in case of Seurat object}

\item{stopAtSwitch}{logical, set to TRUE to stop the attack when the new target
set to FALSE to continue the attack until all genes are tested}

\item{verbose}{logical, set to TRUE to activate verbose mode}
}
\value{
a list containing the modified expression matrix, the list of modified genes,
the summary of the attack by gene, the summary of the attack,
the new cell types predictions and the original cell types predictions
}
\description{
Implementation of the IDG4C algorithm.
}
\details{
this function is an implementation of the IDG4C algorithm
which permits to generate adversarial attacks on a classifier on a given
cluster. The attack is done by modifying the expression of the genes by
gradient descent after having inferred if by the finite difference method.
Two parameters alpha and epsilon are used to control the step size of the
modifications.
}
\examples{
MyClassifier <- function(expr, clusters, target) {
   typePredictions <- as.data.frame(matrix(nrow=nrow(expr), ncol=length(unique(clusters))))
   colnames(typePredictions) <- unique(clusters)
   typePredictions[unique(clusters)[1]] <- c(1,0,0,0)
   typePredictions[unique(clusters)[2]] <- c(0,1,1,1)
   rownames(typePredictions) <- 1:4

   list(prediction="T cell", odd=1,
                     typePredictions=t(typePredictions),
                     cellTypes=c("B cell","T cell","T cell","T cell"))
}
rna_expression <- data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
    CD8B=c(2,2,3,3))
genes <- c("CD4", "CD8A")
clusters_id <- c("B cell","T cell","T cell","T cell")

advCGD(rna_expression, clusters_id, "T cell",
  MyClassifier, genes=genes, verbose=TRUE)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/advSingleGene.R
\name{advSingleGene}
\alias{advSingleGene}
\title{Find a one gene min change adversarial attack list. A one gene min change
adversarial attack  refers to the modification of a single gene
within a cluster, leading to a change in its classification. The function
returns a list of genes/new classification.}
\usage{
advSingleGene(
  exprs,
  clusters,
  target,
  classifier,
  exclGenes = c(),
  genes = c(),
  advMethod = "perc99",
  advFixedValue = 3,
  advFct = NULL,
  firstDichot = 100,
  maxSplitSize = 1,
  returnFirstFound = FALSE,
  changeType = "any",
  argForClassif = "data.frame",
  argForModif = "data.frame",
  slot = NULL,
  verbose = FALSE
)
}
\arguments{
\item{exprs}{DelayedMatrix of numeric RNA expression, cells are rows and genes
are columns - or a SingleCellExperiment object, a matrix or a data.frame. By default,
these are converted to a data.frame to increase speed performance during modifications.
However, this conversion can consume a significant amount of memory, see 'argForModif'
argument for options.}

\item{clusters}{a character vector of the clusters to which the cells belong}

\item{target}{the name of the cluster to modify}

\item{classifier}{a classifier in the suitable format.
A classifier function should be formated as follow:
classifier = function(expr, clusters, target){
# Making the classification
c("cell type", score)
}
\code{score} should be numeric between 0 and 1, 1 being the highest confidance
into the cell type classification.
The matrix \code{expr} contains RNA expression values, the vector \code{clusters}
consists of the cluster IDs for each cell in \code{expr}, and \code{target} is the
ID of the cluster for which we want to have a classification.
The function returns a vector with the classification result, and a score.}

\item{exclGenes}{a character vector of genes to exclude from the analysis}

\item{genes}{a character vector of genes in case you want to limit the
attack on a subset of genes}

\item{advMethod}{the name of the method to use}

\item{advFixedValue}{the numeric value to use in case of
advMethod=\code{fixed}}

\item{advFct}{the function to use in case advMethod
belongs to the following list: \code{full_row_fct}, \code{target_row_fct},
\code{target_matrix_fct}, \code{full_matrix_fct}}

\item{firstDichot}{the initial number of slices before
the dichotomic search}

\item{maxSplitSize}{max size of dichotomic slices}

\item{returnFirstFound}{set to TRUE to return result when a
the first misclassification is found}

\item{changeType}{\code{any} consider each misclassification,
\code{not_na} consider each misclassification but NA.}

\item{argForClassif}{the type of the first argument to feed to the
classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
or 'DelayedMatrix'.}

\item{argForModif}{type of matrix during for the modification, 'data.frame'
by default. Can be 'DelayedMatrix', which needs less memory but is slower.}

\item{verbose}{logical, set to TRUE to activate verbose mode}
}
\value{
a list of genes/new classification tuples
}
\description{
Find a one gene min change adversarial attack list. A one gene min change
adversarial attack  refers to the modification of a single gene
within a cluster, leading to a change in its classification. The function
returns a list of genes/new classification.
}
\details{
This function aims to get all genes that when modified individually
can lead to a misclassification. You can refer to the
'advModifications' function documentation to more details on how
to define a modification.
The function is made as a two step parameter search. The first step is to
split the genes in 'firstDichot' sets, 100 by default. Then each set is
studied by a dichotomic process in a recursive function.
The aim of sarting by a high value of sets, instead of starting directly by
the dichotomic research is to avoid the following scenario: we modify 5000
genes, the modification of one gene conpensates the modification of another.
The classification remains unchanged, whereas there is a one gene
classification modifying inside the 5000.
The dichotomic process runs as follow. The function receives a list of
genes, make the modification of the whole list and make the classification.
Three scenarios possible:
\itemize{
\item the classification remains the same as the target cluster. The function
returns, and the dichotomic process continues.
\item the classification is changed. There is only one gene in the list, the
function returns the gene and the new classification.
\item the classification is changed. There is more than one gene in the list,
the genes list is split in two, and the dichotomic process continues.
}
}
\examples{
library(DelayedArray)

MyClassifier <- function(expr, clusters, target) {
   c("B cell", 0.9)
}
rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
     CD8B=c(2,2,3,3)))
genes <- c("CD4", "CD8A")
clusters_id <- c("B cell","B cell","T cell","T cell")

advSingleGene(rna_expression, clusters_id,
"T cell", MyClassifier, advMethod="perc99")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MClassifier.R
\name{MClassifier}
\alias{MClassifier}
\title{Example cell type classifier for pbmc clustered datasets.}
\usage{
MClassifier(exprs, clusters, target)
}
\arguments{
\item{exprs}{DelayedMatrix of numeric RNA expression, cells are rows and genes
are columns - or a SingleCellExperiment object, a matrix or a data.frame.}

\item{clusters}{vector of clusters to which each cell belongs}

\item{target}{name of the cell cluster to classify}
}
\value{
a vector with the classification, and the odd
}
\description{
Example cell type classifier for pbmc clustered datasets.
}
\details{
This classifier aims at testing the adverSCarial
package of real pbmc data. It is a simple marker based
classifier. It looks at the average value of a few genes
inside a cluster, and returns the associated cell type.
Markers where found by differential expressions.
}
\examples{
library(TENxPBMCData)

pbmc <- TENxPBMCData(dataset = "pbmc3k")
mat_rna <- matrixFromSCE(pbmc)
cell_types <- system.file("extdata",
    "pbmc3k_cell_types.tsv",
    package = "adverSCarial"
)
cell_types <- read.table(cell_types, sep = "\t")$cell_type

MClassifier(mat_rna, cell_types, "DC")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allClasses.R
\docType{class}
\name{advList-class}
\alias{advList-class}
\alias{advList}
\title{adverSCarial class}
\value{
A advList object
}
\description{
\code{advList} is a class used to store the output values
of the \code{advSingleGene} function. The result can be a list
of few thousands genes:cell_type key/values, so a specific \emph{show}
method is associated to this class to avoid overflooding the R
scripts outputs.
}
\examples{
MyClassifier <- function(expr, clusters, target) {
   c("B cell", 0.9)
}
rna_expression <- data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
     CD8B=c(2,2,3,3))
genes <- c("CD4", "CD8A")
clusters_id <- c("B cell","B cell","T cell","T cell")

adv_min_change <- advSingleGene(rna_expression, clusters_id,
"T cell", MyClassifier, advMethod="perc99")

adv_min_change

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/minSingleGeneOverview.R
\name{singleGeneOverview}
\alias{singleGeneOverview}
\title{Gives an overview of the susceptibility to single gene
attacks, for each cell type, for a given list of modifications.}
\usage{
singleGeneOverview(
  exprs,
  clusters,
  classifier,
  exclGenes = c(),
  genes = c(),
  modifications = list(c("perc1"), c("perc99")),
  advMethod = "perc99",
  advFixedValue = 3,
  advFct = NULL,
  firstDichot = 100,
  maxSplitSize = 100,
  changeType = "any",
  argForClassif = "data.frame",
  argForModif = "data.frame",
  verbose = FALSE
)
}
\arguments{
\item{exprs}{DelayedMatrix of numeric RNA expression, cells are rows and genes
are columns - or a SingleCellExperiment object, a matrix or a data.frame. By default,
these are converted to a data.frame to increase speed performance during modifications.
However, this conversion can consume a significant amount of memory, see 'argForModif'
argument for options.}

\item{clusters}{a character vector of the clusters to which the cells belong}

\item{classifier}{a classifier in the suitable format.
A classifier function should be formated as follow:
classifier = function(expr, clusters, target){
# Making the classification
c("cell type", score)
}
\code{score} should be numeric between 0 and 1, 1 being the highest confidance
into the cell type classification.
The matrix \code{expr} contains RNA expression values, the vector \code{clusters}
consists of the cluster IDs for each cell in \code{expr}, and \code{target} is the
ID of the cluster for which we want to have a classification.
The function returns a vector with the classification result, and a score.}

\item{exclGenes}{a character vector of genes to exclude from the analysis}

\item{genes}{a character vector of genes in case you want to limit the
analysis on a subset of genes}

\item{modifications}{the list of the modifications to study}

\item{advMethod}{the name of the method to use}

\item{advFixedValue}{the numeric value to use in case of
advMethod=\code{fixed}}

\item{advFct}{the function to use in case advMethod
belongs to the following list: \code{full_row_fct}, \code{target_row_fct},
\code{target_matrix_fct}, \code{full_matrix_fct}}

\item{firstDichot}{the initial number of slices before
the dichotomic search}

\item{maxSplitSize}{max size of dichotomic slices.}

\item{changeType}{\code{any} consider each misclassification,
\code{not_na} consider each misclassification but NA.}

\item{argForClassif}{the type of the first argument to feed to the
classifier function. 'data.frame' by default, can be 'SingleCellExperiment'
or 'DelayedMatrix'.}

\item{argForModif}{type of matrix during for the modification, 'data.frame'
by default. Can be 'DelayedMatrix', which needs less memory but is slower.}

\item{verbose}{logical, set to TRUE to activate verbose mode}
}
\value{
a DataFrame storing the number of possible single gene
attacks each cell type and each modification.
}
\description{
Gives an overview of the susceptibility to single gene
attacks, for each cell type, for a given list of modifications.
}
\details{
Running the advSingleGene function for each cell type
to see which ones are more vulerable can take a long time. The
aim of the singleGeneOverview function is to make this process faster.
It uses a default value of 100 for the 'maxSplitSize' parameter. So,
the dichotomic process of the advSingleGene function stops as soon
as the fold length is lower than 100. You can have more accurate
results with maxSplitSize=1, but it will take longer.
This function aims also to run the advSingleGene for several given
modifications. You can specify a list of modifications as so - each
item of the list should be 1 or 2 length size.
The 1 length vector must contain the prerecorded modifications, 'perc1'
or 'perc99'.
The 2 length vector must have as first item:
\itemize{
\item 'fixed', in this case the second item should be the value to be
replaced by.
\item 'full_row_fct', 'target_row_fct', 'target_matrix_fct' or
'full_matrix_fct'. In this case the second item should be a function.
Let's say we want to analysis the susceptibility to single gene attack
for 3 modifications: "perc1", the modification of each value of the
cluster by 1000, and a custom modification stored inside a function myFct.
Then the 'modification' parameter should be:
my_modifications = list(c("perc1"),
c("fixed", 1000),
c("full_matrix_fct", myFct))
}

The function returns a dataframe with the number of genes of the max
change attack for each modification in columns, for each cell type in rows.
}
\examples{
library(DelayedArray)

MyClassifier <- function(expr, clusters, target) {
   c("T cell", 0.9)
}
rna_expression <- DelayedArray(data.frame(CD4=c(0,0,0,0), CD8A=c(1,1,1,1),
     CD8B=c(2,2,3,3)))
genes <- c("CD4", "CD8A")
clusters_id <- c("B cell","B cell","T cell","T cell")

singleGeneOverview(rna_expression, clusters_id,
MyClassifier, modifications = list(c("perc1"), c("perc99")))

myModif = function(x, y){
   return(sample(1:10,1))
}

my_modifications = list(c("perc1"),
                        c("fixed", 1000),
                        c("full_matrix_fct", myModif))
singleGeneOverview(rna_expression, clusters_id,
 MyClassifier, modifications = my_modifications)

}
